"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import pytest
import salt.modules.cmdmod
import salt.modules.rpm_lowpkg as rpm
import salt.utils.path
from tests.support.mock import MagicMock, patch

# pylint: disable=unused-import
try:
    import rpm as rpm_lib

    HAS_RPM = True
except ImportError:
    HAS_RPM = False

try:
    import rpm_vercmp

    HAS_PY_RPM = True
except ImportError:
    HAS_PY_RPM = False
# pylint: enable=unused-import


def _called_with_root(mock):
    cmd = " ".join(mock.call_args[0][0])
    return cmd.startswith("rpm --root /")


@pytest.fixture
def configure_loader_modules():
    return {rpm: {"rpm": MagicMock(return_value=MagicMock)}}


# 'list_pkgs' function tests: 2


def test_list_pkgs():
    """
    Test if it list the packages currently installed in a dict
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm.__salt__, {"cmd.run": mock}):
        assert rpm.list_pkgs() == {}
        assert not _called_with_root(mock)


def test_list_pkgs_root():
    """
    Test if it list the packages currently installed in a dict,
    called with root parameter
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm.__salt__, {"cmd.run": mock}):
        rpm.list_pkgs(root="/")
        assert _called_with_root(mock)


# 'verify' function tests: 2


def test_verify():
    """
    Test if it runs an rpm -Va on a system, and returns the
    results in a dict
    """
    mock = MagicMock(
        return_value={"stdout": "", "stderr": "", "retcode": 0, "pid": 12345}
    )
    with patch.dict(rpm.__salt__, {"cmd.run_all": mock}):
        assert rpm.verify("httpd") == {}
        assert not _called_with_root(mock)


def test_verify_root():
    """
    Test if it runs an rpm -Va on a system, and returns the
    results in a dict, called with root parameter
    """
    mock = MagicMock(
        return_value={"stdout": "", "stderr": "", "retcode": 0, "pid": 12345}
    )
    with patch.dict(rpm.__salt__, {"cmd.run_all": mock}):
        rpm.verify("httpd", root="/")
        assert _called_with_root(mock)


# 'file_list' function tests: 2


def test_file_list():
    """
    Test if it list the files that belong to a package.
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm.__salt__, {"cmd.run": mock}):
        assert rpm.file_list("httpd") == {"errors": [], "files": []}
        assert not _called_with_root(mock)


def test_file_list_root():
    """
    Test if it list the files that belong to a package, using the
    root parameter.
    """

    mock = MagicMock(return_value="")
    with patch.dict(rpm.__salt__, {"cmd.run": mock}):
        rpm.file_list("httpd", root="/")
        assert _called_with_root(mock)


# 'file_dict' function tests: 2


def test_file_dict():
    """
    Test if it list the files that belong to a package
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm.__salt__, {"cmd.run": mock}):
        assert rpm.file_dict("httpd") == {"errors": [], "packages": {}}
        assert not _called_with_root(mock)


def test_file_dict_root():
    """
    Test if it list the files that belong to a package
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm.__salt__, {"cmd.run": mock}):
        rpm.file_dict("httpd", root="/")
        assert _called_with_root(mock)


# 'owner' function tests: 1


def test_owner():
    """
    Test if it return the name of the package that owns the file.
    """
    assert rpm.owner() == ""

    ret = "file /usr/bin/salt-jenkins-build is not owned by any package"
    mock = MagicMock(return_value=ret)
    with patch.dict(rpm.__salt__, {"cmd.run_stdout": mock}):
        assert rpm.owner("/usr/bin/salt-jenkins-build") == ""
        assert not _called_with_root(mock)

    ret = {
        "/usr/bin/vim": "vim-enhanced-7.4.160-1.e17.x86_64",
        "/usr/bin/python": "python-2.7.5-16.e17.x86_64",
    }
    mock = MagicMock(
        side_effect=[
            "python-2.7.5-16.e17.x86_64",
            "vim-enhanced-7.4.160-1.e17.x86_64",
        ]
    )
    with patch.dict(rpm.__salt__, {"cmd.run_stdout": mock}):
        assert rpm.owner("/usr/bin/python", "/usr/bin/vim") == ret
        assert not _called_with_root(mock)


def test_owner_root():
    """
    Test if it return the name of the package that owns the file,
    using the parameter root.
    """
    assert rpm.owner() == ""

    ret = "file /usr/bin/salt-jenkins-build is not owned by any package"
    mock = MagicMock(return_value=ret)
    with patch.dict(rpm.__salt__, {"cmd.run_stdout": mock}):
        rpm.owner("/usr/bin/salt-jenkins-build", root="/")
        assert _called_with_root(mock)


# 'checksum' function tests: 2


def test_checksum():
    """
    Test if checksum validate as expected
    """
    ret = {
        "file1.rpm": True,
        "file2.rpm": False,
        "file3.rpm": False,
    }

    mock = MagicMock(side_effect=[True, 0, True, 1, False, 0])
    with patch.dict(rpm.__salt__, {"file.file_exists": mock, "cmd.retcode": mock}):
        assert rpm.checksum("file1.rpm", "file2.rpm", "file3.rpm") == ret
        assert not _called_with_root(mock)


def test_checksum_root():
    """
    Test if checksum validate as expected, using the parameter
    root
    """
    mock = MagicMock(side_effect=[True, 0])
    with patch.dict(rpm.__salt__, {"file.file_exists": mock, "cmd.retcode": mock}):
        rpm.checksum("file1.rpm", root="/")
        assert _called_with_root(mock)


@pytest.mark.parametrize("rpm_lib", ["HAS_RPM", "HAS_PY_RPM", "rpmdev-vercmp"])
def test_version_cmp_rpm_all_libraries(rpm_lib):
    """
    Test package version when each library is installed
    """
    rpmdev = salt.utils.path.which("rpmdev-vercmp")
    patch_cmd = patch.dict(rpm.__salt__, {"cmd.run_all": salt.modules.cmdmod.run_all})
    if rpm_lib == "rpmdev-vercmp":
        if rpmdev:
            patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", False)
            patch_py_rpm = patch("salt.modules.rpm_lowpkg.HAS_PY_RPM", False)
        else:
            pytest.skip("The rpmdev-vercmp binary is not installed")
    elif rpm_lib == "HAS_RPM":
        if HAS_RPM:
            patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", True)
            patch_py_rpm = patch("salt.modules.rpm_lowpkg.HAS_PY_RPM", False)
        else:
            pytest.skip("The RPM lib is not installed, skipping")
    elif rpm_lib == "HAS_PY_RPM":
        if HAS_PY_RPM:
            patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", False)
            patch_py_rpm = patch("salt.modules.rpm_lowpkg.HAS_PY_RPM", True)
        else:
            pytest.skip("The Python RPM lib is not installed, skipping")

    with patch_rpm, patch_py_rpm, patch_cmd:
        assert -1 == rpm.version_cmp("1", "2")
        assert -1 == rpm.version_cmp("2.9.1-6.el7_2.3", "2.9.1-6.el7.4")
        assert 1 == rpm.version_cmp("3.2", "3.0")
        assert 0 == rpm.version_cmp("3.0", "3.0")
        assert 1 == rpm.version_cmp("1:2.9.1-6.el7_2.3", "2.9.1-6.el7.4")
        assert -1 == rpm.version_cmp("1:2.9.1-6.el7_2.3", "1:2.9.1-6.el7.4")
        assert 1 == rpm.version_cmp("2:2.9.1-6.el7_2.3", "1:2.9.1-6.el7.4")
        assert 0 == rpm.version_cmp("3:2.9.1-6.el7.4", "3:2.9.1-6.el7.4")
        assert -1 == rpm.version_cmp("3:2.9.1-6.el7.4", "3:2.9.1-7.el7.4")
        assert 1 == rpm.version_cmp("3:2.9.1-8.el7.4", "3:2.9.1-7.el7.4")


def test_version_cmp_rpm():
    """
    Test package version if RPM-Python is installed

    :return:
    """
    mock_label = MagicMock(return_value=-1)
    mock_log = MagicMock()
    patch_label = patch("salt.modules.rpm_lowpkg.rpm.labelCompare", mock_label)
    patch_log = patch("salt.modules.rpm_lowpkg.log", mock_log)
    patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", True)
    with patch_label, patch_rpm, patch_log:
        assert -1 == rpm.version_cmp("1", "2")
        assert not mock_log.warning.called
        assert mock_label.called


def test_version_cmp_rpmutils():
    """
    Test package version if rpmUtils.miscutils called

    :return:
    """
    mock_log = MagicMock()
    mock_rpmUtils = MagicMock()
    mock_rpmUtils.miscutils = MagicMock()
    mock_rpmUtils.miscutils.compareEVR = MagicMock(return_value=-1)
    patch_utils = patch("salt.modules.rpm_lowpkg.rpmUtils", mock_rpmUtils, create=True)
    patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", False)
    patch_utils_lib = patch("salt.modules.rpm_lowpkg.HAS_RPMUTILS", True)
    patch_py_rpm = patch("salt.modules.rpm_lowpkg.HAS_PY_RPM", False)
    patch_log = patch("salt.modules.rpm_lowpkg.log", mock_log)

    with patch_utils, patch_rpm, patch_py_rpm, patch_utils_lib, patch_log:
        assert -1 == rpm.version_cmp("1", "2")
        assert mock_log.warning.called
        assert mock_rpmUtils.miscutils.compareEVR.called


def test_version_cmp_rpmdev_vercmp():
    """
    Test package version if rpmdev-vercmp is installed

    :return:
    """
    mock__salt__ = MagicMock(return_value={"retcode": 12})
    mock_log = MagicMock()
    patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", False)
    patch_rpmutils = patch("salt.modules.rpm_lowpkg.HAS_RPMUTILS", False)
    patch_py_rpm = patch("salt.modules.rpm_lowpkg.HAS_PY_RPM", False)
    patch_which = patch("salt.utils.path.which", return_value=True)
    patch_log = patch("salt.modules.rpm_lowpkg.log", mock_log)

    with patch_rpm, patch_rpmutils, patch_py_rpm, patch_which, patch_log:
        with patch.dict(rpm.__salt__, {"cmd.run_all": mock__salt__}):
            assert -1 == rpm.version_cmp("1", "2")
            assert mock__salt__.called
            assert mock_log.warning.called
            assert (
                mock_log.warning.mock_calls[0][1][0]
                == "Please install a package that provides rpm.labelCompare for more accurate version comparisons."
            )
            assert (
                mock_log.warning.mock_calls[1][1][0]
                == "Installing the rpmdevtools package may surface dev tools in production."
            )


def test_version_cmp_python():
    """
    Test package version if falling back to python

    :return:
    """
    mock_log = MagicMock()
    patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", False)
    patch_rpmutils = patch("salt.modules.rpm_lowpkg.HAS_RPMUTILS", False)
    mock_version_cmp = MagicMock(return_value=-1)
    patch_py_rpm = patch("salt.modules.rpm_lowpkg.HAS_PY_RPM", False)
    patch_cmp = patch("salt.utils.versions.version_cmp", mock_version_cmp)
    patch_which = patch("salt.utils.path.which", return_value=False)
    patch_log = patch("salt.modules.rpm_lowpkg.log", mock_log)

    with patch_rpm, patch_rpmutils, patch_py_rpm, patch_cmp, patch_which, patch_log:
        assert -1 == rpm.version_cmp("1", "2")
        assert mock_version_cmp.called
        assert mock_log.warning.called
        assert (
            mock_log.warning.mock_calls[0][1][0]
            == "Please install a package that provides rpm.labelCompare for more accurate version comparisons."
        )
        assert (
            mock_log.warning.mock_calls[1][1][0]
            == "Falling back on salt.utils.versions.version_cmp() for version comparisons"
        )



"""
Integration tests for the rabbitmq_user states
"""

import logging

import pytest
import salt.modules.rabbitmq as rabbitmq
import salt.states.rabbitmq_user as rabbitmq_user

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_freebsd(reason="No Docker on FreeBSD available"),
    pytest.mark.skip_if_binaries_missing(
        "docker", "dockerd", reason="Docker not installed"
    ),
]


@pytest.fixture
def configure_loader_modules(docker_cmd_run_all_wrapper):
    return {
        rabbitmq_user: {
            "__salt__": {
                "rabbitmq.user_exists": rabbitmq.user_exists,
                "rabbitmq.add_user": rabbitmq.add_user,
                "rabbitmq.delete_user": rabbitmq.delete_user,
                "rabbitmq.list_user_permissions": rabbitmq.list_user_permissions,
            },
            "__opts__": {"test": False},
            "_utils__": {},
        },
        rabbitmq: {
            "__salt__": {
                "rabbitmq.user_exists": rabbitmq.user_exists,
                "rabbitmq.add_user": rabbitmq.add_user,
                "rabbitmq.list_user_permissions": rabbitmq.list_user_permissions,
                "cmd.run_all": docker_cmd_run_all_wrapper,
            },
            "__opts__": {},
            "_utils__": {},
        },
    }


@pytest.mark.slow_test
def test_present_absent(docker_cmd_run_all_wrapper, rabbitmq_container):
    """
    Test rabbitmq_user.present

    Create user then delete it.
    """

    # Clear the user
    ret = rabbitmq_user.present("myuser")
    expected = {
        "name": "myuser",
        "result": True,
        "comment": "'myuser' was configured.",
        "changes": {"user": {"old": "", "new": "myuser"}},
    }
    assert ret == expected

    # Delete the user
    ret = rabbitmq_user.absent("myuser")
    expected = {
        "name": "myuser",
        "result": True,
        "comment": "The user 'myuser' was removed.",
        "changes": {"name": {"old": "myuser", "new": ""}},
    }

    assert ret == expected


def test_absent(docker_cmd_run_all_wrapper, rabbitmq_container):
    """
    Test rabbitmq_user.absent
    """

    # Delete the user
    ret = rabbitmq_user.absent("myuser")
    expected = {
        "name": "myuser",
        "result": True,
        "comment": "The user 'myuser' is not present.",
        "changes": {},
    }
    assert ret == expected



import os
import shutil

import pytest
import salt.utils.files
import salt.utils.user
from tests.support.case import ModuleCase
from tests.support.mixins import AdaptedConfigurationTestCaseMixin
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf

# Doesn't work. Why?
# @requires_salt_modules('acl')
# @requires_salt_modules('linux_acl')
# Acl package should be installed to test linux_acl module


@pytest.mark.skip_if_binaries_missing("getfacl")
class LinuxAclModuleTest(ModuleCase, AdaptedConfigurationTestCaseMixin):
    """
    Validate the linux_acl module
    """

    def setUp(self):
        # Blindly copied from tests.integration.modules.file; Refactoring?
        self.myfile = os.path.join(RUNTIME_VARS.TMP, "myfile")
        with salt.utils.files.fopen(self.myfile, "w+") as fp:
            fp.write("Hello\n")
        self.mydir = os.path.join(RUNTIME_VARS.TMP, "mydir/isawesome")
        if not os.path.isdir(self.mydir):
            # left behind... Don't fail because of this!
            os.makedirs(self.mydir)
        self.mysymlink = os.path.join(RUNTIME_VARS.TMP, "mysymlink")
        if os.path.islink(self.mysymlink):
            os.remove(self.mysymlink)
        os.symlink(self.myfile, self.mysymlink)
        self.mybadsymlink = os.path.join(RUNTIME_VARS.TMP, "mybadsymlink")
        if os.path.islink(self.mybadsymlink):
            os.remove(self.mybadsymlink)
        os.symlink("/nonexistentpath", self.mybadsymlink)
        super().setUp()

    def tearDown(self):
        if os.path.isfile(self.myfile):
            os.remove(self.myfile)
        if os.path.islink(self.mysymlink):
            os.remove(self.mysymlink)
        if os.path.islink(self.mybadsymlink):
            os.remove(self.mybadsymlink)
        shutil.rmtree(self.mydir, ignore_errors=True)
        super().tearDown()

    @skipIf(salt.utils.platform.is_freebsd(), "Skip on FreeBSD")
    def test_version(self):
        self.assertRegex(self.run_function("acl.version"), r"\d+\.\d+\.\d+")

    @skipIf(salt.utils.platform.is_freebsd(), "Skip on FreeBSD")
    def test_getfacl_w_single_file_without_acl(self):
        ret = self.run_function("acl.getfacl", arg=[self.myfile])
        user = salt.utils.user.get_user()
        group = salt.utils.user.get_default_group(user)
        self.maxDiff = None
        self.assertEqual(
            ret,
            {
                self.myfile: {
                    "other": [
                        {
                            "": {
                                "octal": 4,
                                "permissions": {
                                    "read": True,
                                    "write": False,
                                    "execute": False,
                                },
                            }
                        }
                    ],
                    "user": [
                        {
                            user: {
                                "octal": 6,
                                "permissions": {
                                    "read": True,
                                    "write": True,
                                    "execute": False,
                                },
                            }
                        }
                    ],
                    "group": [
                        {
                            group: {
                                "octal": 4,
                                "permissions": {
                                    "read": True,
                                    "write": False,
                                    "execute": False,
                                },
                            }
                        }
                    ],
                    "comment": {"owner": user, "group": group, "file": self.myfile},
                }
            },
        )



import copy
import os

import salt.modules.grains as grainsmod
import salt.utils.dictupdate as dictupdate
from salt.exceptions import SaltException
from salt.utils.odict import OrderedDict
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class GrainsModuleTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        conf_file = os.path.join(RUNTIME_VARS.TMP, "__salt_test_grains")
        cachedir = os.path.join(RUNTIME_VARS.TMP, "__salt_test_grains_cache_dir")
        if not os.path.isdir(cachedir):
            os.makedirs(cachedir)
        return {
            grainsmod: {
                "__opts__": {"conf_file": conf_file, "cachedir": cachedir},
                "__salt__": {"saltutil.refresh_grains": MagicMock()},
            }
        }

    def test_filter_by(self):
        with patch.dict(
            grainsmod.__grains__,
            {"os_family": "MockedOS", "1": "1", "2": "2", "roles": ["A", "B"]},
        ):

            dict1 = {"A": "B", "C": {"D": {"E": "F", "G": "H"}}}
            dict2 = {
                "default": {"A": "B", "C": {"D": "E"}},
                "1": {"A": "X"},
                "2": {"C": {"D": "H"}},
                "MockedOS": {"A": "Z"},
            }

            mdict1 = {"D": {"E": "I"}, "J": "K"}
            mdict2 = {"A": "Z"}
            mdict3 = {"C": {"D": "J"}}

            # test None result with non existent grain and no default
            res = grainsmod.filter_by(dict1, grain="xxx")
            self.assertIs(res, None)

            # test None result with os_family grain and no matching result
            res = grainsmod.filter_by(dict1)
            self.assertIs(res, None)

            # test with non existent grain, and a given default key
            res = grainsmod.filter_by(dict1, grain="xxx", default="C")
            self.assertEqual(res, {"D": {"E": "F", "G": "H"}})

            # add a merge dictionary, F disappears
            res = grainsmod.filter_by(dict1, grain="xxx", merge=mdict1, default="C")
            self.assertEqual(res, {"D": {"E": "I", "G": "H"}, "J": "K"})
            # dict1 was altered, reestablish
            dict1 = {"A": "B", "C": {"D": {"E": "F", "G": "H"}}}

            # default is not present in dict1, check we only have merge in result
            res = grainsmod.filter_by(dict1, grain="xxx", merge=mdict1, default="Z")
            self.assertEqual(res, mdict1)

            # default is not present in dict1, and no merge, should get None
            res = grainsmod.filter_by(dict1, grain="xxx", default="Z")
            self.assertIs(res, None)

            # test giving a list as merge argument raise exception
            self.assertRaises(
                SaltException, grainsmod.filter_by, dict1, "xxx", ["foo"], "C"
            )

            # Now, re-test with an existing grain (os_family), but with no match.
            res = grainsmod.filter_by(dict1)
            self.assertIs(res, None)
            res = grainsmod.filter_by(dict1, default="C")
            self.assertEqual(res, {"D": {"E": "F", "G": "H"}})
            res = grainsmod.filter_by(dict1, merge=mdict1, default="C")
            self.assertEqual(res, {"D": {"E": "I", "G": "H"}, "J": "K"})
            # dict1 was altered, reestablish
            dict1 = {"A": "B", "C": {"D": {"E": "F", "G": "H"}}}
            res = grainsmod.filter_by(dict1, merge=mdict1, default="Z")
            self.assertEqual(res, mdict1)
            res = grainsmod.filter_by(dict1, default="Z")
            self.assertIs(res, None)
            # this one is in fact a traceback in updatedict, merging a string with a dictionary
            self.assertRaises(
                TypeError, grainsmod.filter_by, dict1, merge=mdict1, default="A"
            )

            # Now, re-test with a matching grain.
            dict1 = {"A": "B", "MockedOS": {"D": {"E": "F", "G": "H"}}}
            res = grainsmod.filter_by(dict1)
            self.assertEqual(res, {"D": {"E": "F", "G": "H"}})
            res = grainsmod.filter_by(dict1, default="A")
            self.assertEqual(res, {"D": {"E": "F", "G": "H"}})
            res = grainsmod.filter_by(dict1, merge=mdict1, default="A")
            self.assertEqual(res, {"D": {"E": "I", "G": "H"}, "J": "K"})
            # dict1 was altered, reestablish
            dict1 = {"A": "B", "MockedOS": {"D": {"E": "F", "G": "H"}}}
            res = grainsmod.filter_by(dict1, merge=mdict1, default="Z")
            self.assertEqual(res, {"D": {"E": "I", "G": "H"}, "J": "K"})
            # dict1 was altered, reestablish
            dict1 = {"A": "B", "MockedOS": {"D": {"E": "F", "G": "H"}}}
            res = grainsmod.filter_by(dict1, default="Z")
            self.assertEqual(res, {"D": {"E": "F", "G": "H"}})

            # Test when grain value is a list
            dict1 = {"A": "B", "C": {"D": {"E": "F", "G": "H"}}}
            res = grainsmod.filter_by(dict1, grain="roles", default="C")
            self.assertEqual(res, "B")
            # Test default when grain value is a list
            dict1 = {"Z": "B", "C": {"D": {"E": "F", "G": "H"}}}
            res = grainsmod.filter_by(dict1, grain="roles", default="C")
            self.assertEqual(res, {"D": {"E": "F", "G": "H"}})

            # Test with wildcard pattern in the lookup_dict keys
            dict1 = {"*OS": "B", "C": {"D": {"E": "F", "G": "H"}}}
            res = grainsmod.filter_by(dict1)
            self.assertEqual(res, "B")
            # Test with non-strings in lookup_dict keys
            # Issue #38094
            dict1 = {1: 2, 3: {4: 5}, "*OS": "B"}
            res = grainsmod.filter_by(dict1)
            self.assertEqual(res, "B")
            # Test with sequence pattern with roles
            dict1 = {"Z": "B", "[BC]": {"D": {"E": "F", "G": "H"}}}
            res = grainsmod.filter_by(dict1, grain="roles", default="Z")
            self.assertEqual(res, {"D": {"E": "F", "G": "H"}})

            # Base tests
            # NOTE: these may fail to detect errors if dictupdate.update() is broken
            # but then the unit test for dictupdate.update() should fail and expose
            # that.  The purpose of these tests is it validate the logic of how
            # in filter_by() processes its arguments.

            # Test with just the base
            res = grainsmod.filter_by(dict2, grain="xxx", default="xxx", base="default")
            self.assertEqual(res, dict2["default"])

            # Test the base with the OS grain look-up
            res = grainsmod.filter_by(dict2, default="xxx", base="default")
            self.assertEqual(
                res,
                dictupdate.update(copy.deepcopy(dict2["default"]), dict2["MockedOS"]),
            )

            # Test the base with default
            res = grainsmod.filter_by(dict2, grain="xxx", base="default")
            self.assertEqual(res, dict2["default"])

            res = grainsmod.filter_by(dict2, grain="1", base="default")
            self.assertEqual(
                res, dictupdate.update(copy.deepcopy(dict2["default"]), dict2["1"])
            )

            res = grainsmod.filter_by(dict2, base="default", merge=mdict2)
            self.assertEqual(
                res,
                dictupdate.update(
                    dictupdate.update(
                        copy.deepcopy(dict2["default"]), dict2["MockedOS"]
                    ),
                    mdict2,
                ),
            )

            res = grainsmod.filter_by(dict2, base="default", merge=mdict3)
            self.assertEqual(
                res,
                dictupdate.update(
                    dictupdate.update(
                        copy.deepcopy(dict2["default"]), dict2["MockedOS"]
                    ),
                    mdict3,
                ),
            )

    def test_append_not_a_list(self):
        # Failing append to an existing string, without convert
        with patch.dict(grainsmod.__grains__, {"b": "bval"}):
            res = grainsmod.append("b", "d")
            self.assertEqual(res, "The key b is not a valid list")
            self.assertEqual(grainsmod.__grains__, {"b": "bval"})

            # Failing append to an existing dict
            with patch.dict(grainsmod.__grains__, {"b": {"b1": "bval1"}}):
                res = grainsmod.append("b", "d")
                self.assertEqual(res, "The key b is not a valid list")
                self.assertEqual(grainsmod.__grains__, {"b": {"b1": "bval1"}})

    def test_append_already_in_list(self):
        # Append an existing value
        with patch.dict(grainsmod.__grains__, {"a_list": ["a", "b", "c"], "b": "bval"}):
            res = grainsmod.append("a_list", "b")
            self.assertEqual(res, "The val b was already in the list a_list")
            self.assertEqual(
                grainsmod.__grains__, {"a_list": ["a", "b", "c"], "b": "bval"}
            )

    def test_append_ok(self):
        # Append to an existing list
        with patch.dict(grainsmod.__grains__, {"a_list": ["a", "b", "c"], "b": "bval"}):
            res = grainsmod.append("a_list", "d")
            self.assertEqual(res, {"a_list": ["a", "b", "c", "d"]})
            self.assertEqual(
                grainsmod.__grains__, {"a_list": ["a", "b", "c", "d"], "b": "bval"}
            )

        # Append to an non existing list
        with patch.dict(grainsmod.__grains__, {"b": "bval"}):
            res = grainsmod.append("a_list", "d")
            self.assertEqual(res, {"a_list": ["d"]})
            self.assertEqual(grainsmod.__grains__, {"a_list": ["d"], "b": "bval"})

        # Append to an existing string, with convert
        with patch.dict(grainsmod.__grains__, {"b": "bval"}):
            res = grainsmod.append("b", "d", convert=True)
            self.assertEqual(res, {"b": ["bval", "d"]})
            self.assertEqual(grainsmod.__grains__, {"b": ["bval", "d"]})

        # Append to an existing dict, with convert
        with patch.dict(grainsmod.__grains__, {"b": {"b1": "bval1"}}):
            res = grainsmod.append("b", "d", convert=True)
            self.assertEqual(res, {"b": [{"b1": "bval1"}, "d"]})
            self.assertEqual(grainsmod.__grains__, {"b": [{"b1": "bval1"}, "d"]})

    def test_append_nested_not_a_list(self):
        # Failing append to an existing string, without convert
        with patch.dict(grainsmod.__grains__, {"a": {"b": "bval"}}):
            res = grainsmod.append("a:b", "d")
            self.assertEqual(res, "The key a:b is not a valid list")
            self.assertEqual(grainsmod.__grains__, {"a": {"b": "bval"}})

        # Failing append to an existing dict
        with patch.dict(grainsmod.__grains__, {"a": {"b": {"b1": "bval1"}}}):
            res = grainsmod.append("a:b", "d")
            self.assertEqual(res, "The key a:b is not a valid list")
            self.assertEqual(grainsmod.__grains__, {"a": {"b": {"b1": "bval1"}}})

    def test_append_nested_already_in_list(self):
        # Append an existing value
        with patch.dict(
            grainsmod.__grains__, {"a": {"a_list": ["a", "b", "c"], "b": "bval"}}
        ):
            res = grainsmod.append("a:a_list", "b")
            self.assertEqual(res, "The val b was already in the list a:a_list")
            self.assertEqual(
                grainsmod.__grains__, {"a": {"a_list": ["a", "b", "c"], "b": "bval"}}
            )

    def test_append_nested_ok(self):
        # Append to an existing list
        with patch.dict(
            grainsmod.__grains__, {"a": {"a_list": ["a", "b", "c"], "b": "bval"}}
        ):
            res = grainsmod.append("a:a_list", "d")
            self.assertEqual(res, {"a": {"a_list": ["a", "b", "c", "d"], "b": "bval"}})
            self.assertEqual(
                grainsmod.__grains__,
                {"a": {"a_list": ["a", "b", "c", "d"], "b": "bval"}},
            )

        # Append to an non existing list
        with patch.dict(grainsmod.__grains__, {"a": {"b": "bval"}}):
            res = grainsmod.append("a:a_list", "d")
            self.assertEqual(res, {"a": {"a_list": ["d"], "b": "bval"}})
            self.assertEqual(
                grainsmod.__grains__, {"a": {"a_list": ["d"], "b": "bval"}}
            )

        # Append to an existing string, with convert
        with patch.dict(grainsmod.__grains__, {"a": {"b": "bval"}}):
            res = grainsmod.append("a:b", "d", convert=True)
            self.assertEqual(res, {"a": {"b": ["bval", "d"]}})
            self.assertEqual(grainsmod.__grains__, {"a": {"b": ["bval", "d"]}})

        # Append to an existing dict, with convert
        with patch.dict(grainsmod.__grains__, {"a": {"b": {"b1": "bval1"}}}):
            res = grainsmod.append("a:b", "d", convert=True)
            self.assertEqual(res, {"a": {"b": [{"b1": "bval1"}, "d"]}})
            self.assertEqual(grainsmod.__grains__, {"a": {"b": [{"b1": "bval1"}, "d"]}})

    def test_append_to_an_element_of_a_list(self):
        # Append to an element in a list
        # It currently fails silently
        with patch.dict(grainsmod.__grains__, {"a": ["b", "c"]}):
            res = grainsmod.append("a:b", "d")
            self.assertEqual(res, {"a": ["b", "c"]})
            self.assertEqual(grainsmod.__grains__, {"a": ["b", "c"]})

    def test_set_value_already_set(self):
        # Set a grain to the same simple value
        with patch.dict(grainsmod.__grains__, {"a": 12, "c": 8}):
            res = grainsmod.set("a", 12)
            self.assertTrue(res["result"])
            self.assertEqual(res["comment"], "Grain is already set")
            self.assertEqual(grainsmod.__grains__, {"a": 12, "c": 8})

        # Set a grain to the same complex value
        with patch.dict(grainsmod.__grains__, {"a": ["item", 12], "c": 8}):
            res = grainsmod.set("a", ["item", 12])
            self.assertTrue(res["result"])
            self.assertEqual(res["comment"], "Grain is already set")
            self.assertEqual(grainsmod.__grains__, {"a": ["item", 12], "c": 8})

        # Set a key to the same simple value in a nested grain
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
        ):
            res = grainsmod.set("b,nested", "val", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(res["comment"], "Grain is already set")
            self.assertEqual(
                grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
            )

    def test_set_fail_replacing_existing_complex_key(self):
        # Fails to set a complex value without 'force'
        with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
            res = grainsmod.set("a", ["item", 12])
            self.assertFalse(res["result"])
            self.assertEqual(
                res["comment"],
                "The key 'a' exists and the given value is a "
                "dict or a list. Use 'force=True' to overwrite.",
            )
            self.assertEqual(grainsmod.__grains__, {"a": "aval", "c": 8})

        # Fails to overwrite a complex value without 'force'
        with patch.dict(grainsmod.__grains__, {"a": ["item", 12], "c": 8}):
            res = grainsmod.set("a", ["item", 14])
            self.assertFalse(res["result"])
            self.assertEqual(
                res["comment"],
                "The key 'a' exists but is a dict or a list. "
                + "Use 'force=True' to overwrite.",
            )
            self.assertEqual(grainsmod.__grains__, {"a": ["item", 12], "c": 8})

        # Fails to overwrite a complex value without 'force' in a nested grain
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": ["val1"]}], "c": 8}
        ):
            res = grainsmod.set("b,l2", "val2", delimiter=",")
            self.assertFalse(res["result"])
            self.assertEqual(
                res["comment"],
                "The key 'b,l2' exists but is a dict or a "
                + "list. Use 'force=True' to overwrite.",
            )
            self.assertEqual(
                grainsmod.__grains__,
                {"a": "aval", "b": ["l1", {"l2": ["val1"]}], "c": 8},
            )

    def test_set_nested_fails_replace_simple_value(self):
        # Fails to replace a simple value with a new dictionary consisting
        # of the specified key and value
        with patch.dict(grainsmod.__grains__, {"a": "aval", "b": "l1", "c": 8}):
            res = grainsmod.set("b,l3", "val3", delimiter=",")
            self.assertFalse(res["result"])
            self.assertEqual(
                res["comment"],
                "The key 'b' value is 'l1', which is "
                + "different from the provided key 'l3'. "
                + "Use 'force=True' to overwrite.",
            )
            self.assertEqual(grainsmod.__grains__, {"a": "aval", "b": "l1", "c": 8})

    def test_set_simple_value(self):
        with patch.dict(grainsmod.__grains__, {"a": ["b", "c"], "c": 8}):
            res = grainsmod.set("b", "bval")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": "bval"})
            self.assertEqual(
                grainsmod.__grains__, {"a": ["b", "c"], "b": "bval", "c": 8}
            )

    def test_set_replace_value(self):
        with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
            res = grainsmod.set("a", 12)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"a": 12})
            self.assertEqual(grainsmod.__grains__, {"a": 12, "c": 8})

    def test_set_None_ok(self):
        with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
            res = grainsmod.set("b", None)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": None})
            self.assertEqual(grainsmod.__grains__, {"a": "aval", "b": None, "c": 8})

    def test_set_None_ok_destructive(self):
        with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
            res = grainsmod.set("b", None, destructive=True)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": None})
            self.assertEqual(grainsmod.__grains__, {"a": "aval", "c": 8})

    def test_set_None_replace_ok(self):
        with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
            res = grainsmod.set("a", None)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"a": None})
            self.assertEqual(grainsmod.__grains__, {"a": None, "c": 8})

    def test_set_None_force_destructive(self):
        with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
            res = grainsmod.set("a", None, force=True, destructive=True)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"a": None})
            self.assertEqual(grainsmod.__grains__, {"c": 8})

    def test_set_replace_value_was_complex_force(self):
        with patch.dict(grainsmod.__grains__, {"a": ["item", 12], "c": 8}):
            res = grainsmod.set("a", "aval", force=True)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"a": "aval"})
            self.assertEqual(grainsmod.__grains__, {"a": "aval", "c": 8})

    def test_set_complex_value_force(self):
        with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
            res = grainsmod.set("a", ["item", 12], force=True)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"a": ["item", 12]})
            self.assertEqual(grainsmod.__grains__, {"a": ["item", 12], "c": 8})

    def test_set_nested_create(self):
        with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
            res = grainsmod.set("b,nested", "val", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": {"nested": "val"}})
            self.assertEqual(
                grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
            )

    def test_set_nested_update_dict(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
        ):
            res = grainsmod.set("b,nested", "val2", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": {"nested": "val2"}})
            self.assertEqual(
                grainsmod.__grains__, {"a": "aval", "b": {"nested": "val2"}, "c": 8}
            )

    def test_set_nested_update_dict_remove_key(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
        ):
            res = grainsmod.set("b,nested", None, delimiter=",", destructive=True)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": {}})
            self.assertEqual(grainsmod.__grains__, {"a": "aval", "b": {}, "c": 8})

    def test_set_nested_update_dict_new_key(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
        ):
            res = grainsmod.set("b,b2", "val2", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": {"b2": "val2", "nested": "val"}})
            self.assertEqual(
                grainsmod.__grains__,
                {"a": "aval", "b": {"b2": "val2", "nested": "val"}, "c": 8},
            )

    def test_set_nested_list_replace_key(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": ["l1", "l2", "l3"], "c": 8}
        ):
            res = grainsmod.set("b,l2", "val2", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": ["l1", {"l2": "val2"}, "l3"]})
            self.assertEqual(
                grainsmod.__grains__,
                {"a": "aval", "b": ["l1", {"l2": "val2"}, "l3"], "c": 8},
            )

    def test_set_nested_list_update_dict_key(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": "val1"}], "c": 8}
        ):
            res = grainsmod.set("b,l2", "val2", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": ["l1", {"l2": "val2"}]})
            self.assertEqual(
                grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": "val2"}], "c": 8}
            )

    def test_set_nested_list_update_dict_key_overwrite(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": ["val1"]}], "c": 8}
        ):
            res = grainsmod.set("b,l2", "val2", delimiter=",", force=True)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": ["l1", {"l2": "val2"}]})
            self.assertEqual(
                grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": "val2"}], "c": 8}
            )

    def test_set_nested_list_append_dict_key(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": "val2"}], "c": 8}
        ):
            res = grainsmod.set("b,l3", "val3", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(
                res["changes"], {"b": ["l1", {"l2": "val2"}, {"l3": "val3"}]}
            )
            self.assertEqual(
                grainsmod.__grains__,
                {"a": "aval", "b": ["l1", {"l2": "val2"}, {"l3": "val3"}], "c": 8},
            )

    def test_set_nested_existing_value_is_the_key(self):
        with patch.dict(grainsmod.__grains__, {"a": "aval", "b": "l3", "c": 8}):
            res = grainsmod.set("b,l3", "val3", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": {"l3": "val3"}})
            self.assertEqual(
                grainsmod.__grains__, {"a": "aval", "b": {"l3": "val3"}, "c": 8}
            )

    def test_set_nested_existing_value_overwrite(self):
        with patch.dict(grainsmod.__grains__, {"a": "aval", "b": "l1", "c": 8}):
            res = grainsmod.set("b,l3", "val3", delimiter=",", force=True)
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": {"l3": "val3"}})
            self.assertEqual(
                grainsmod.__grains__, {"a": "aval", "b": {"l3": "val3"}, "c": 8}
            )

    def test_set_deeply_nested_update(self):
        with patch.dict(
            grainsmod.__grains__,
            {"a": "aval", "b": {"l1": ["l21", "l22", {"l23": "l23val"}]}, "c": 8},
        ):
            res = grainsmod.set("b,l1,l23", "val", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(
                res["changes"], {"b": {"l1": ["l21", "l22", {"l23": "val"}]}}
            )
            self.assertEqual(
                grainsmod.__grains__,
                {"a": "aval", "b": {"l1": ["l21", "l22", {"l23": "val"}]}, "c": 8},
            )

    def test_set_deeply_nested_create(self):
        with patch.dict(
            grainsmod.__grains__,
            {"a": "aval", "b": {"l1": ["l21", "l22", {"l23": "l23val"}]}, "c": 8},
        ):
            res = grainsmod.set("b,l1,l24,l241", "val", delimiter=",")
            self.assertTrue(res["result"])
            self.assertEqual(
                res["changes"],
                {
                    "b": {
                        "l1": [
                            "l21",
                            "l22",
                            {"l23": "l23val"},
                            {"l24": {"l241": "val"}},
                        ]
                    }
                },
            )
            self.assertEqual(
                grainsmod.__grains__,
                {
                    "a": "aval",
                    "b": {
                        "l1": [
                            "l21",
                            "l22",
                            {"l23": "l23val"},
                            {"l24": {"l241": "val"}},
                        ]
                    },
                    "c": 8,
                },
            )

    def test_get_ordered(self):
        with patch.dict(
            grainsmod.__grains__,
            OrderedDict(
                [
                    ("a", "aval"),
                    (
                        "b",
                        OrderedDict(
                            [
                                ("z", "zval"),
                                (
                                    "l1",
                                    ["l21", "l22", OrderedDict([("l23", "l23val")])],
                                ),
                            ]
                        ),
                    ),
                    ("c", 8),
                ]
            ),
        ):
            res = grainsmod.get("b")
            self.assertEqual(type(res), OrderedDict)
            # Check that order really matters
            self.assertTrue(
                res
                == OrderedDict(
                    [
                        ("z", "zval"),
                        ("l1", ["l21", "l22", OrderedDict([("l23", "l23val")])]),
                    ]
                )
            )
            self.assertFalse(
                res
                == OrderedDict(
                    [
                        ("l1", ["l21", "l22", OrderedDict([("l23", "l23val")])]),
                        ("z", "zval"),
                    ]
                )
            )

    def test_get_unordered(self):
        with patch.dict(
            grainsmod.__grains__,
            OrderedDict(
                [
                    ("a", "aval"),
                    (
                        "b",
                        OrderedDict(
                            [
                                ("z", "zval"),
                                (
                                    "l1",
                                    ["l21", "l22", OrderedDict([("l23", "l23val")])],
                                ),
                            ]
                        ),
                    ),
                    ("c", 8),
                ]
            ),
        ):
            res = grainsmod.get("b", ordered=False)
            self.assertEqual(type(res), dict)
            # Check that order doesn't matter
            self.assertTrue(
                res
                == OrderedDict(
                    [
                        ("l1", ["l21", "l22", OrderedDict([("l23", "l23val")])]),
                        ("z", "zval"),
                    ]
                )
            )

    def test_equals(self):
        with patch.dict(
            grainsmod.__grains__,
            OrderedDict(
                [
                    ("a", "aval"),
                    (
                        "b",
                        OrderedDict(
                            [
                                ("z", "zval"),
                                (
                                    "l1",
                                    ["l21", "l22", OrderedDict([("l23", "l23val")])],
                                ),
                            ]
                        ),
                    ),
                    ("c", 8),
                ]
            ),
        ):
            res = grainsmod.equals("a", "aval")
            self.assertEqual(type(res), bool)
            self.assertTrue(res)
            res = grainsmod.equals("b:z", "zval")
            self.assertTrue(res)
            res = grainsmod.equals("b:z", "aval")
            self.assertFalse(res)

    def test_grains_setval_refresh_pillar(self):
        """
        Test that refresh_pillar kwarg is being passed correctly from grains.setval to saltutil.refresh_grains
        """
        ret = grainsmod.setval("test_grains_setval_refresh_pillar", "saltopotamus")
        grainsmod.__salt__["saltutil.refresh_grains"].assert_called_with(
            refresh_pillar=True
        )
        ret = grainsmod.setval(
            "test_grains_setval_refresh_pillar", "saltopotamus", refresh_pillar=True
        )
        grainsmod.__salt__["saltutil.refresh_grains"].assert_called_with(
            refresh_pillar=True
        )
        ret = grainsmod.setval(
            "test_grains_setval_refresh_pillar", "saltopotamus", refresh_pillar=False
        )
        grainsmod.__salt__["saltutil.refresh_grains"].assert_called_with(
            refresh_pillar=False
        )

    def test_setval_unicode(self):
        key = "塩"  # salt
        value = "塩人生です"  # salt is life

        # Note: call setvals 2 times is important
        # 1: add key to conf grains
        # 2: update and read key from conf grains
        for _ in range(2):
            ret = grainsmod.setvals({key: value})
            self.assertIn(key, ret)
            self.assertEqual(ret[key], value)

    def test_delval_single(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
        ):
            res = grainsmod.delval("a")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"a": None})
            self.assertEqual(
                grainsmod.__grains__, {"a": None, "b": {"nested": "val"}, "c": 8}
            )

    def test_delval_nested(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
        ):
            res = grainsmod.delval("b:nested")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": {"nested": None}})
            self.assertEqual(
                grainsmod.__grains__, {"a": "aval", "b": {"nested": None}, "c": 8}
            )

    def test_delkey_single_key(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
        ):
            res = grainsmod.delkey("a")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"a": None})
            self.assertEqual(grainsmod.__grains__, {"b": {"nested": "val"}, "c": 8})

    def test_delkey_nested_key(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
        ):
            res = grainsmod.delkey("b:nested")
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": {}})
            self.assertEqual(grainsmod.__grains__, {"a": "aval", "b": {}, "c": 8})

    def test_delkey_nested_key_force_needed(self):
        with patch.dict(
            grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
        ):
            res = grainsmod.delkey("b", force=True)
            assert res["comment"].find("Use 'force=True' to overwrite.") == -1
            self.assertTrue(res["result"])
            self.assertEqual(res["changes"], {"b": None})
            self.assertEqual(grainsmod.__grains__, {"a": "aval", "c": 8})



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    tests.unit.utils.vt_test
    ~~~~~~~~~~~~~~~~~~~~~~~~

    VirtualTerminal tests
"""


import functools
import io
import os
import random
import subprocess
import sys
import time

import salt.utils
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.vt
from tests.support.paths import CODE_DIR
from tests.support.unit import TestCase, skipIf


def stdout_fileno_available():
    """
    Tests if sys.stdout.fileno is available in this testing environment
    """
    try:
        sys.stdout.fileno()
        return True
    except io.UnsupportedOperation:
        return False


def fixStdOutErrFileNoIfNeeded(func):
    """
    Decorator that sets stdout and stderr to their original objects if
    sys.stdout.fileno() doesn't work and restores them after running the
    decorated function. This doesn't check if the original objects actually
    work. If they don't then the test environment is too broken to test
    the VT.
    """

    @functools.wraps(func)
    def wrapper_fixStdOutErrFileNoIfNeeded(*args, **kwargs):
        original_stdout = os.sys.stdout
        original_stderr = os.sys.stderr
        if not stdout_fileno_available():
            os.sys.stdout = os.sys.__stdout__
            os.sys.stderr = os.sys.__stderr__
        try:
            return func(*args, **kwargs)
        finally:
            os.sys.stdout = original_stdout
            os.sys.stderr = original_stderr

    return wrapper_fixStdOutErrFileNoIfNeeded


class VTTestCase(TestCase):
    @skipIf(
        salt.utils.platform.is_windows(),
        "Skip on Windows because this feature is not supported",
    )
    def test_vt_size(self):
        """Confirm that the terminal size is being set"""
        cols = random.choice(range(80, 250))
        terminal = salt.utils.vt.Terminal(
            "stty size",
            shell=True,
            cols=cols,
            rows=24,
            stream_stdout=False,
            stream_stderr=False,
        )
        self.assertEqual(terminal.getwinsize(), (24, cols))
        buffer_o = buffer_e = ""
        while terminal.has_unread_data:
            stdout, stderr = terminal.recv()
            if stdout:
                buffer_o += stdout
            if stderr:
                buffer_e += stderr
        assert buffer_o.strip() == "24 {}".format(cols)
        try:
            # Then wait for the terminal child to exit, this will raise an
            # exception if the process has already exited.
            terminal.wait()
        except salt.utils.vt.TerminalException:
            pass
        terminal.close()

    @skipIf(
        True,
        "Disabled until we can find out why this kills the tests suite with an exit"
        " code of 134",
    )
    def test_issue_10404_ptys_not_released(self):
        n_executions = 15

        def current_pty_count():
            # Get current number of PTY's
            try:
                if os.path.exists("/proc/sys/kernel/pty/nr"):
                    with salt.utils.files.fopen("/proc/sys/kernel/pty/nr") as fh_:
                        return int(fh_.read().strip())

                proc = subprocess.Popen(
                    "sysctl -a 2> /dev/null | grep pty.nr | awk '{print $3}'",
                    shell=True,
                    stdout=subprocess.PIPE,
                )
                stdout, _ = proc.communicate()
                return int(stdout.strip())
            except (ValueError, OSError):
                if salt.utils.platform.is_darwin():
                    # We're unable to findout how many PTY's are open
                    self.skipTest(
                        "Unable to find out how many PTY's are open on Darwin - "
                        "Skipping for now"
                    )
                self.fail("Unable to find out how many PTY's are open")

        nr_ptys = current_pty_count()

        # Using context manager's
        for idx in range(0, nr_ptys + n_executions):
            try:
                with salt.utils.vt.Terminal(
                    'echo "Run {}"'.format(idx),
                    shell=True,
                    stream_stdout=False,
                    stream_stderr=False,
                ) as terminal:
                    terminal.wait()
                try:
                    if current_pty_count() > (nr_ptys + (n_executions / 2)):
                        self.fail("VT is not cleaning up PTY's")
                except (ValueError, OSError):
                    self.fail("Unable to find out how many PTY's are open")
            except Exception as exc:  # pylint: disable=broad-except
                if "out of pty devices" in str(exc):
                    # We're not cleaning up
                    raise
                # We're pushing the system resources, let's keep going
                continue

        # Not using context manager's
        for idx in range(0, nr_ptys + n_executions):
            try:
                terminal = salt.utils.vt.Terminal(
                    'echo "Run {}"'.format(idx),
                    shell=True,
                    stream_stdout=False,
                    stream_stderr=False,
                )
                terminal.wait()
                try:
                    if current_pty_count() > (nr_ptys + (n_executions / 2)):
                        self.fail("VT is not cleaning up PTY's")
                except (ValueError, OSError):
                    self.fail("Unable to find out how many PTY's are open")
            except Exception as exc:  # pylint: disable=broad-except
                if "out of pty devices" in str(exc):
                    # We're not cleaning up
                    raise
                # We're pushing the system resources, let's keep going
                continue

    @skipIf(True, "Disabled until we can figure out how to make this more reliable.")
    def test_isalive_while_theres_data_to_read(self):
        expected_data = "Alive!\n"
        term = salt.utils.vt.Terminal(
            'echo "Alive!"', shell=True, stream_stdout=False, stream_stderr=False
        )
        buffer_o = buffer_e = ""
        try:
            while term.has_unread_data:
                stdout, stderr = term.recv()
                if stdout:
                    buffer_o += stdout
                if stderr:
                    buffer_e += stderr
                # While there's data to be read, the process is alive
                if stdout is None and stderr is None:
                    self.assertFalse(term.isalive())

            # term should be dead now
            self.assertEqual(buffer_o, expected_data)
            self.assertFalse(term.isalive())

            stdout, stderr = term.recv()
            self.assertFalse(term.isalive())
            self.assertIsNone(stderr)
            self.assertIsNone(stdout)
        finally:
            term.close(terminate=True, kill=True)

        expected_data = "Alive!\n"
        term = salt.utils.vt.Terminal(
            'echo "Alive!" 1>&2', shell=True, stream_stdout=False, stream_stderr=False
        )
        buffer_o = buffer_e = ""
        try:
            while term.has_unread_data:
                stdout, stderr = term.recv()
                if stdout:
                    buffer_o += stdout
                if stderr:
                    buffer_e += stderr
                # While there's data to be read, the process is alive
                if stdout is None and stderr is None:
                    self.assertFalse(term.isalive())

            # term should be dead now
            self.assertEqual(buffer_e, expected_data)
            self.assertFalse(term.isalive())

            stdout, stderr = term.recv()
            self.assertFalse(term.isalive())
            self.assertIsNone(stderr)
            self.assertIsNone(stdout)
        finally:
            term.close(terminate=True, kill=True)

        expected_data = "Alive!\nAlive!\n"
        term = salt.utils.vt.Terminal(
            'echo "Alive!"; sleep 5; echo "Alive!"',
            shell=True,
            stream_stdout=False,
            stream_stderr=False,
        )
        buffer_o = buffer_e = ""
        try:
            while term.has_unread_data:
                stdout, stderr = term.recv()
                if stdout:
                    buffer_o += stdout
                if stderr:
                    buffer_e += stderr
                # While there's data to be read, the process is alive
                if stdout is None and stderr is None:
                    self.assertFalse(term.isalive())

                if buffer_o != expected_data:
                    self.assertTrue(term.isalive())
                # Don't spin
                time.sleep(0.1)

            # term should be dead now
            self.assertEqual(buffer_o, expected_data)
            self.assertFalse(term.isalive())

            stdout, stderr = term.recv()
            self.assertFalse(term.isalive())
            self.assertIsNone(stderr)
            self.assertIsNone(stdout)
        finally:
            term.close(terminate=True, kill=True)

    @staticmethod
    def generate_multibyte_stdout_unicode(block_size):
        return b"\xE2\x80\xA6" * 4 * block_size

    @staticmethod
    def generate_multibyte_stderr_unicode(block_size):
        return b"\x2E" + VTTestCase.generate_multibyte_stdout_unicode(block_size)

    @skipIf(
        salt.utils.platform.is_windows(), "Skip VT tests on windows, due to issue 54290"
    )
    @fixStdOutErrFileNoIfNeeded
    def test_split_multibyte_characters_unicode(self):
        """
        Tests that the vt correctly handles multibyte characters that are
        split between blocks of transmitted data.
        """
        block_size = 1024
        encoding = "utf-8"
        stdout_content = VTTestCase.generate_multibyte_stdout_unicode(block_size)
        # stderr is offset by one byte to guarentee a split character in
        # one of the output streams
        stderr_content = VTTestCase.generate_multibyte_stderr_unicode(block_size)

        expected_stdout = salt.utils.stringutils.to_unicode(stdout_content, encoding)
        expected_stderr = salt.utils.stringutils.to_unicode(stderr_content, encoding)
        python_command = "\n".join(
            (
                "import sys",
                "import os",
                "import tests.unit.utils.test_vt as test_vt",
                (
                    "os.write(sys.stdout.fileno(), "
                    "test_vt.VTTestCase.generate_multibyte_stdout_unicode("
                    + str(block_size)
                    + "))"
                ),
                (
                    "os.write(sys.stderr.fileno(), "
                    "test_vt.VTTestCase.generate_multibyte_stderr_unicode("
                    + str(block_size)
                    + "))"
                ),
            )
        )
        term = salt.utils.vt.Terminal(
            args=[sys.executable, "-c", '"' + python_command + '"'],
            shell=True,
            cwd=CODE_DIR,
            stream_stdout=False,
            stream_stderr=False,
            force_receive_encoding=encoding,
        )
        buffer_o = buffer_e = salt.utils.stringutils.to_unicode("")
        try:
            while term.has_unread_data:
                stdout, stderr = term.recv(block_size)
                if stdout:
                    buffer_o += stdout
                if stderr:
                    buffer_e += stderr

            self.assertEqual(buffer_o, expected_stdout)
            self.assertEqual(buffer_e, expected_stderr)
        finally:
            term.close(terminate=True, kill=True)

    @staticmethod
    def generate_multibyte_stdout_shiftjis(block_size):
        return b"\x8B\x80" * 4 * block_size

    @staticmethod
    def generate_multibyte_stderr_shiftjis(block_size):
        return b"\x2E" + VTTestCase.generate_multibyte_stdout_shiftjis(block_size)

    @skipIf(
        salt.utils.platform.is_windows(), "Skip VT tests on windows, due to issue 54290"
    )
    @fixStdOutErrFileNoIfNeeded
    def test_split_multibyte_characters_shiftjis(self):
        """
        Tests that the vt correctly handles multibyte characters that are
        split between blocks of transmitted data.
        Uses shift-jis encoding to make sure code doesn't assume unicode.
        """
        block_size = 1024
        encoding = "shift-jis"
        stdout_content = VTTestCase.generate_multibyte_stdout_shiftjis(block_size)
        stderr_content = VTTestCase.generate_multibyte_stderr_shiftjis(block_size)

        expected_stdout = salt.utils.stringutils.to_unicode(stdout_content, encoding)
        expected_stderr = salt.utils.stringutils.to_unicode(stderr_content, encoding)
        python_command = "\n".join(
            (
                "import sys",
                "import os",
                "import tests.unit.utils.test_vt as test_vt",
                (
                    "os.write(sys.stdout.fileno(), "
                    "test_vt.VTTestCase.generate_multibyte_stdout_shiftjis("
                    + str(block_size)
                    + "))"
                ),
                (
                    "os.write(sys.stderr.fileno(), "
                    "test_vt.VTTestCase.generate_multibyte_stderr_shiftjis("
                    + str(block_size)
                    + "))"
                ),
            )
        )
        term = salt.utils.vt.Terminal(
            args=[sys.executable, "-c", '"' + python_command + '"'],
            shell=True,
            cwd=CODE_DIR,
            stream_stdout=False,
            stream_stderr=False,
            force_receive_encoding=encoding,
        )
        buffer_o = buffer_e = salt.utils.stringutils.to_unicode("")
        try:
            while term.has_unread_data:
                stdout, stderr = term.recv(block_size)
                if stdout:
                    buffer_o += stdout
                if stderr:
                    buffer_e += stderr

            self.assertEqual(buffer_o, expected_stdout)
            self.assertEqual(buffer_e, expected_stderr)
        finally:
            term.close(terminate=True, kill=True)



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import salt.modules.saltcloudmod as saltcloudmod
import salt.utils.json
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SaltcloudmodTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.saltcloudmod
    """

    def setup_loader_modules(self):
        return {saltcloudmod: {}}

    def setUp(self):
        self.mock_json_loads = MagicMock(side_effect=ValueError())

    def test_create(self):
        """
        Test if create the named vm
        """
        mock = MagicMock(return_value="""{"foo": "bar"}""")
        with patch.dict(saltcloudmod.__salt__, {"cmd.run_stdout": mock}):
            self.assertTrue(saltcloudmod.create("webserver", "rackspace_centos_512"))

            with patch.object(salt.utils.json, "loads", self.mock_json_loads):
                self.assertDictEqual(
                    saltcloudmod.create("webserver", "rackspace_centos_512"), {}
                )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.rsync as rsync
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class RsyncTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.rsync
    """

    def setup_loader_modules(self):
        return {rsync: {}}

    def test_rsync(self):
        """
        Test for rsync files from src to dst
        """
        with patch.dict(
            rsync.__salt__, {"config.option": MagicMock(return_value=False)}
        ):
            self.assertRaises(SaltInvocationError, rsync.rsync, "", "")

        with patch.dict(
            rsync.__salt__,
            {
                "config.option": MagicMock(return_value="A"),
                "cmd.run_all": MagicMock(side_effect=[OSError(1, "f"), "A"]),
            },
        ):
            with patch.object(rsync, "_check", return_value=["A"]):
                self.assertRaises(CommandExecutionError, rsync.rsync, "a", "b")

                self.assertEqual(rsync.rsync("src", "dst"), "A")

    def test_version(self):
        """
        Test for return rsync version
        """
        mock = MagicMock(side_effect=[OSError(1, "f"), "A B C\n"])
        with patch.dict(rsync.__salt__, {"cmd.run_stdout": mock}):
            self.assertRaises(CommandExecutionError, rsync.version)

            self.assertEqual(rsync.version(), "C")

    def test_rsync_excludes_list(self):
        """
        Test for rsync files from src to dst with a list of excludes
        """
        mock = {
            "config.option": MagicMock(return_value=False),
            "cmd.run_all": MagicMock(),
        }
        with patch.dict(rsync.__salt__, mock):
            rsync.rsync("src", "dst", exclude=["test/one", "test/two"])
        mock["cmd.run_all"].assert_called_once_with(
            [
                "rsync",
                "-avz",
                "--exclude",
                "test/one",
                "--exclude",
                "test/two",
                "src",
                "dst",
            ],
            python_shell=False,
        )

    def test_rsync_excludes_str(self):
        """
        Test for rsync files from src to dst with one exclude
        """
        mock = {
            "config.option": MagicMock(return_value=False),
            "cmd.run_all": MagicMock(),
        }
        with patch.dict(rsync.__salt__, mock):
            rsync.rsync("src", "dst", exclude="test/one")
        mock["cmd.run_all"].assert_called_once_with(
            ["rsync", "-avz", "--exclude", "test/one", "src", "dst"],
            python_shell=False,
        )



import salt.modules.mac_xattr as xattr
import salt.utils.mac_utils
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class XAttrTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {xattr: {}}

    def test_list(self):
        """
        Test xattr.list
        """
        expected = {"spongebob": "squarepants", "squidward": "patrick"}
        with patch.object(
            xattr, "read", MagicMock(side_effect=["squarepants", "patrick"])
        ), patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value="spongebob\nsquidward"),
        ):
            self.assertEqual(xattr.list_("path/to/file"), expected)

    def test_list_missing(self):
        """
        Test listing attributes of a missing file
        """
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(side_effect=CommandExecutionError("No such file")),
        ):
            self.assertRaises(CommandExecutionError, xattr.list_, "/path/to/file")

    def test_read(self):
        """
        Test reading a specific attribute from a file
        """
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(return_value="expected results"),
        ):
            self.assertEqual(
                xattr.read("/path/to/file", "com.attr"), "expected results"
            )

    def test_read_hex(self):
        """
        Test reading a specific attribute from a file
        """
        with patch.object(
            salt.utils.mac_utils,
            "execute_return_result",
            MagicMock(return_value="expected results"),
        ) as mock:
            self.assertEqual(
                xattr.read("/path/to/file", "com.attr", **{"hex": True}),
                "expected results",
            )
            mock.assert_called_once_with(
                ["xattr", "-p", "-x", "com.attr", "/path/to/file"]
            )

    def test_read_missing(self):
        """
        Test reading a specific attribute from a file
        """
        with patch(
            "salt.utils.mac_utils.execute_return_result",
            MagicMock(side_effect=CommandExecutionError("No such file")),
        ):
            self.assertRaises(
                CommandExecutionError, xattr.read, "/path/to/file", "attribute"
            )

    def test_write(self):
        """
        Test writing a specific attribute to a file
        """
        mock_cmd = MagicMock(return_value="squarepants")
        with patch.object(xattr, "read", mock_cmd), patch(
            "salt.utils.mac_utils.execute_return_success", MagicMock(return_value=True)
        ):
            self.assertTrue(xattr.write("/path/to/file", "spongebob", "squarepants"))

    def test_write_missing(self):
        """
        Test writing a specific attribute to a file
        """
        with patch(
            "salt.utils.mac_utils.execute_return_success",
            MagicMock(side_effect=CommandExecutionError("No such file")),
        ):
            self.assertRaises(
                CommandExecutionError,
                xattr.write,
                "/path/to/file",
                "attribute",
                "value",
            )

    def test_delete(self):
        """
        Test deleting a specific attribute from a file
        """
        mock_cmd = MagicMock(return_value={"spongebob": "squarepants"})
        with patch.object(xattr, "list_", mock_cmd), patch(
            "salt.utils.mac_utils.execute_return_success", MagicMock(return_value=True)
        ):
            self.assertTrue(xattr.delete("/path/to/file", "attribute"))

    def test_delete_missing(self):
        """
        Test deleting a specific attribute from a file
        """
        with patch(
            "salt.utils.mac_utils.execute_return_success",
            MagicMock(side_effect=CommandExecutionError("No such file")),
        ):
            self.assertRaises(
                CommandExecutionError, xattr.delete, "/path/to/file", "attribute"
            )

    def test_clear(self):
        """
        Test clearing all attributes on a file
        """
        mock_cmd = MagicMock(return_value={})
        with patch.object(xattr, "list_", mock_cmd), patch(
            "salt.utils.mac_utils.execute_return_success", MagicMock(return_value=True)
        ):
            self.assertTrue(xattr.clear("/path/to/file"))

    def test_clear_missing(self):
        """
        Test clearing all attributes on a file
        """
        with patch(
            "salt.utils.mac_utils.execute_return_success",
            MagicMock(side_effect=CommandExecutionError("No such file")),
        ):
            self.assertRaises(CommandExecutionError, xattr.clear, "/path/to/file")



"""
    :codeauthor: Alexander Schwartz <alexander.schwartz@gmx.net>
"""

import salt.client
from salt.cloud.clouds import saltify
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import ANY, MagicMock, patch
from tests.support.unit import TestCase

TEST_PROFILES = {
    "testprofile1": NotImplemented,
    "testprofile2": {  # this profile is used in test_saltify_destroy()
        "ssh_username": "fred",
        "remove_config_on_destroy": False,  # expected for test
        "shutdown_on_destroy": True,  # expected value for test
    },
    "testprofile3": {  # this profile is used in test_create_wake_on_lan()
        "wake_on_lan_mac": "aa-bb-cc-dd-ee-ff",
        "wol_sender_node": "friend1",
        "wol_boot_wait": 0.01,  # we want the wait to be very short
    },
}
TEST_PROFILE_NAMES = ["testprofile1", "testprofile2", "testprofile3"]


class SaltifyTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.cloud.clouds.saltify
    """

    LOCAL_OPTS = {
        "providers": {
            "sfy1": {"saltify": {"driver": "saltify", "profiles": TEST_PROFILES}},
        },
        "profiles": TEST_PROFILES,
        "sock_dir": "/var/sockxxx",
        "transport": "tcp",
    }

    def setup_loader_modules(self):
        saltify_globals = {
            "__active_provider_name__": "",
            "__utils__": {
                "cloud.bootstrap": MagicMock(),
                "cloud.fire_event": MagicMock(),
            },
            "__opts__": self.LOCAL_OPTS,
        }
        return {saltify: saltify_globals}

    def test_create_no_deploy(self):
        """
        Test if deployment fails. This is the most basic test as saltify doesn't contain much logic
        """
        with patch("salt.cloud.clouds.saltify._verify", MagicMock(return_value=True)):
            vm = {"deploy": False, "driver": "saltify", "name": "dummy"}
            self.assertTrue(saltify.create(vm))

    def test_create_and_deploy(self):
        """
        Test if deployment can be done.
        """
        mock_cmd = MagicMock(return_value=True)
        with patch.dict(
            "salt.cloud.clouds.saltify.__utils__", {"cloud.bootstrap": mock_cmd}
        ):
            vm_ = {
                "deploy": True,
                "driver": "saltify",
                "name": "new2",
                "profile": "testprofile2",
            }
            result = saltify.create(vm_)
            mock_cmd.assert_called_once_with(vm_, ANY)
            self.assertTrue(result)

    def test_create_no_ssh_host(self):
        """
        Test that ssh_host is set to the vm name if not defined
        """
        mock_cmd = MagicMock(return_value=True)
        with patch.dict(
            "salt.cloud.clouds.saltify.__utils__", {"cloud.bootstrap": mock_cmd}
        ):
            vm_ = {
                "deploy": True,
                "driver": "saltify",
                "name": "new2",
                "profile": "testprofile2",
            }
            result = saltify.create(vm_)
            mock_cmd.assert_called_once_with(vm_, ANY)
            assert result
            # Make sure that ssh_host was added to the vm. Note that this is
            # done in two asserts so that the failure is more explicit about
            # what is wrong. If ssh_host wasn't inserted in the vm_ dict, the
            # failure would be a KeyError, which would be harder to
            # troubleshoot.
            assert "ssh_host" in vm_
            assert vm_["ssh_host"] == "new2"

    def test_create_wake_on_lan(self):
        """
        Test if wake on lan works
        """
        mock_sleep = MagicMock()
        mock_cmd = MagicMock(return_value=True)
        mm_cmd = MagicMock(return_value={"friend1": True})
        with salt.client.LocalClient() as lcl:
            lcl.cmd = mm_cmd
            with patch("time.sleep", mock_sleep):
                with patch("salt.client.LocalClient", return_value=lcl):
                    with patch.dict(
                        "salt.cloud.clouds.saltify.__utils__",
                        {"cloud.bootstrap": mock_cmd},
                    ):
                        vm_ = {
                            "deploy": True,
                            "driver": "saltify",
                            "name": "new1",
                            "profile": "testprofile3",
                        }
                        result = saltify.create(vm_)
                        mock_cmd.assert_called_once_with(vm_, ANY)
                        mm_cmd.assert_called_with(
                            "friend1", "network.wol", ["aa-bb-cc-dd-ee-ff"]
                        )
                        # The test suite might call time.sleep, look for any call
                        # that has the expected wait time.
                        mock_sleep.assert_any_call(0.01)
                        self.assertTrue(result)

    def test_avail_locations(self):
        """
        Test the avail_locations will always return {}
        """
        self.assertEqual(saltify.avail_locations(), {})

    def test_avail_sizes(self):
        """
        Test the avail_sizes will always return {}
        """
        self.assertEqual(saltify.avail_sizes(), {})

    def test_avail_images(self):
        """
        Test the avail_images will return profiles
        """
        testlist = list(TEST_PROFILE_NAMES)  # copy
        self.assertEqual(saltify.avail_images()["Profiles"].sort(), testlist.sort())

    def test_list_nodes(self):
        """
        Test list_nodes will return required fields only
        """
        testgrains = {
            "nodeX1": {
                "id": "nodeX1",
                "ipv4": ["127.0.0.1", "192.1.2.22", "172.16.17.18"],
                "ipv6": ["::1", "fdef:bad:add::f00", "3001:DB8::F00D"],
                "salt-cloud": {
                    "driver": "saltify",
                    "provider": "saltyfy",
                    "profile": "testprofile2",
                },
                "extra_stuff": "does not belong",
            }
        }
        expected_result = {
            "nodeX1": {
                "id": "nodeX1",
                "image": "testprofile2",
                "private_ips": ["172.16.17.18", "fdef:bad:add::f00"],
                "public_ips": ["192.1.2.22", "3001:DB8::F00D"],
                "size": "",
                "state": "running",
            }
        }
        mm_cmd = MagicMock(return_value=testgrains)
        with salt.client.LocalClient() as lcl:
            lcl.cmd = mm_cmd
            with patch("salt.client.LocalClient", return_value=lcl):
                self.assertEqual(saltify.list_nodes(), expected_result)

    def test_saltify_reboot(self):
        mm_cmd = MagicMock(return_value=True)
        with salt.client.LocalClient() as lcl:
            lcl.cmd = mm_cmd
            with patch("salt.client.LocalClient", return_value=lcl):
                result = saltify.reboot("nodeS1", "action")
                mm_cmd.assert_called_with("nodeS1", "system.reboot")
                self.assertTrue(result)

    def test_saltify_destroy(self):
        # destroy calls local.cmd several times and expects
        # different results, so we will provide a list of
        # results. Each call will get the next value.
        # NOTE: this assumes that the call order never changes,
        # so to keep things simple, we will not use remove_config...
        result_list = [
            {
                "nodeS1": {  # first call is grains.get
                    "driver": "saltify",
                    "provider": "saltify",
                    "profile": "testprofile2",
                }
            },
            #  Note:
            #    testprofile2 has remove_config_on_destroy: False
            #    and shutdown_on_destroy: True
            {
                "nodeS1": (  # last call shuts down the minion
                    "a system.shutdown worked message"
                )
            },
        ]
        mm_cmd = MagicMock(side_effect=result_list)
        with salt.client.LocalClient() as lcl:
            lcl.cmd = mm_cmd
            with patch("salt.client.LocalClient", return_value=lcl):
                result = saltify.destroy("nodeS1", "action")
                mm_cmd.assert_called_with("nodeS1", "system.shutdown")
                self.assertTrue(result)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.pecl as pecl
from tests.support.mock import patch
from tests.support.unit import TestCase


class PeclTestCase(TestCase):
    """
    Test cases for salt.modules.pecl
    """

    def test_install(self):
        """
        Test to installs one or several pecl extensions.
        """
        with patch.object(pecl, "_pecl", return_value="A"):
            self.assertEqual(pecl.install("fuse", force=True), "A")

            self.assertFalse(pecl.install("fuse"))

            with patch.object(pecl, "list_", return_value={"A": ["A", "B"]}):
                self.assertTrue(pecl.install(["A", "B"]))

    def test_uninstall(self):
        """
        Test to uninstall one or several pecl extensions.
        """
        with patch.object(pecl, "_pecl", return_value="A"):
            self.assertEqual(pecl.uninstall("fuse"), "A")

    def test_update(self):
        """
        Test to update one or several pecl extensions.
        """
        with patch.object(pecl, "_pecl", return_value="A"):
            self.assertEqual(pecl.update("fuse"), "A")

    def test_list_(self):
        """
        Test to list installed pecl extensions.
        """
        with patch.object(pecl, "_pecl", return_value="A\nB"):
            self.assertDictEqual(pecl.list_("channel"), {})



import logging
import random
import string

import salt.config
import salt.loader
import salt.modules.boto_iot as boto_iot
from salt.utils.versions import LooseVersion
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

# pylint: disable=import-error,no-name-in-module,unused-import
try:
    import boto
    import boto3
    from botocore.exceptions import ClientError
    from botocore import __version__ as found_botocore_version

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

# pylint: enable=import-error,no-name-in-module,unused-import

# the boto_iot module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto3_version = "1.2.1"
required_botocore_version = "1.4.41"

log = logging.getLogger(__name__)


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif LooseVersion(boto3.__version__) < LooseVersion(required_boto3_version):
        return False
    elif LooseVersion(found_botocore_version) < LooseVersion(required_botocore_version):
        return False
    else:
        return True


if _has_required_boto():
    region = "us-east-1"
    access_key = "GKTADJGHEIQSXMKKRBJ08H"
    secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
    conn_parameters = {
        "region": region,
        "key": access_key,
        "keyid": secret_key,
        "profile": {},
    }
    error_message = (
        "An error occurred (101) when calling the {0} operation: Test-defined error"
    )
    not_found_error = ClientError(
        {
            "Error": {
                "Code": "ResourceNotFoundException",
                "Message": "Test-defined error",
            }
        },
        "msg",
    )
    topic_rule_not_found_error = ClientError(
        {"Error": {"Code": "UnauthorizedException", "Message": "Test-defined error"}},
        "msg",
    )
    error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}
    policy_ret = dict(
        policyName="testpolicy",
        policyDocument=(
            '{"Version": "2012-10-17", "Statement": [{"Action": ["iot:Publish"],'
            ' "Resource": ["*"], "Effect": "Allow"}]}'
        ),
        policyArn="arn:aws:iot:us-east-1:123456:policy/my_policy",
        policyVersionId=1,
        defaultVersionId=1,
    )
    topic_rule_ret = dict(
        ruleName="testrule",
        sql="SELECT * FROM 'iot/test'",
        description="topic rule description",
        createdAt="1970-01-01",
        actions=[{"lambda": {"functionArn": "arn:aws:::function"}}],
        ruleDisabled=True,
    )

    thing_type_name = "test_thing_type"
    thing_type_desc = "test_thing_type_desc"
    thing_type_attr_1 = "test_thing_type_search_attr_1"
    thing_type_ret = dict(
        thingTypeName=thing_type_name,
        thingTypeProperties=dict(
            thingTypeDescription=thing_type_desc,
            searchableAttributes=[thing_type_attr_1],
        ),
        thingTypeMetadata=dict(
            deprecated=False, creationDate="test_thing_type_create_date"
        ),
    )
    thing_type_arn = "test_thing_type_arn"
    create_thing_type_ret = dict(
        thingTypeName=thing_type_name, thingTypeArn=thing_type_arn
    )


@skipIf(HAS_BOTO is False, "The boto module must be installed.")
@skipIf(
    _has_required_boto() is False,
    "The boto3 module must be greater than or equal to version {}".format(
        required_boto3_version
    ),
)
class BotoIoTTestCaseBase(TestCase, LoaderModuleMockMixin):
    conn = None

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(
            opts, whitelist=["boto3", "args", "systemd", "path", "platform"], context={}
        )
        return {boto_iot: {"__utils__": utils}}

    def setUp(self):
        super().setUp()
        boto_iot.__init__(self.opts)
        del self.opts

        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        self.addCleanup(delattr, self, "conn")
        session_instance.client.return_value = self.conn


class BotoIoTTestCaseMixin:
    pass


class BotoIoTThingTypeTestCase(BotoIoTTestCaseBase, BotoIoTTestCaseMixin):
    """
    TestCase for salt.modules.boto_iot module
    """

    def test_that_when_checking_if_a_thing_type_exists_and_a_thing_type_exists_the_thing_type_exists_method_returns_true(
        self,
    ):
        """
        Tests checking iot thing type existence when the iot thing type already exists
        """
        self.conn.describe_thing_type.return_value = thing_type_ret
        result = boto_iot.thing_type_exists(
            thingTypeName=thing_type_name, **conn_parameters
        )

        self.assertTrue(result["exists"])

    def test_that_when_checking_if_a_thing_type_exists_and_a_thing_type_does_not_exist_the_thing_type_exists_method_returns_false(
        self,
    ):
        """
        Tests checking iot thing type existence when the iot thing type does not exist
        """
        self.conn.describe_thing_type.side_effect = not_found_error
        result = boto_iot.thing_type_exists(
            thingTypeName="non existent thing type", **conn_parameters
        )

        self.assertFalse(result["exists"])

    def test_that_when_checking_if_a_thing_type_exists_and_boto3_returns_an_error_the_thing_type_exists_method_returns_error(
        self,
    ):
        """
        Tests checking iot thing type existence when boto returns an error
        """
        self.conn.describe_thing_type.side_effect = ClientError(
            error_content, "describe_thing_type"
        )
        result = boto_iot.thing_type_exists(
            thingTypeName="mythingtype", **conn_parameters
        )

        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("describe_thing_type"),
        )

    def test_that_when_describing_thing_type_and_thing_type_exists_the_describe_thing_type_method_returns_thing_type(
        self,
    ):
        """
        Tests describe thing type for an existing thing type
        """
        self.conn.describe_thing_type.return_value = thing_type_ret
        result = boto_iot.describe_thing_type(
            thingTypeName=thing_type_name, **conn_parameters
        )

        self.assertEqual(result.get("thing_type"), thing_type_ret)

    def test_that_when_describing_thing_type_and_thing_type_does_not_exists_the_describe_thing_type_method_returns_none(
        self,
    ):
        """
        Tests describe thing type for an non existent thing type
        """
        self.conn.describe_thing_type.side_effect = not_found_error
        result = boto_iot.describe_thing_type(
            thingTypeName="non existent thing type", **conn_parameters
        )

        self.assertEqual(result.get("thing_type"), None)

    def test_that_when_describing_thing_type_and_boto3_returns_error_an_error_the_describe_thing_type_method_returns_error(
        self,
    ):
        self.conn.describe_thing_type.side_effect = ClientError(
            error_content, "describe_thing_type"
        )
        result = boto_iot.describe_thing_type(
            thingTypeName="mythingtype", **conn_parameters
        )

        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("describe_thing_type"),
        )

    def test_that_when_creating_a_thing_type_succeeds_the_create_thing_type_method_returns_true(
        self,
    ):
        """
        tests True when thing type created
        """
        self.conn.create_thing_type.return_value = create_thing_type_ret
        result = boto_iot.create_thing_type(
            thingTypeName=thing_type_name,
            thingTypeDescription=thing_type_desc,
            searchableAttributesList=[thing_type_attr_1],
            **conn_parameters
        )
        self.assertTrue(result["created"])
        self.assertTrue(result["thingTypeArn"], thing_type_arn)

    def test_that_when_creating_a_thing_type_fails_the_create_thing_type_method_returns_error(
        self,
    ):
        """
        tests False when thing type not created
        """
        self.conn.create_thing_type.side_effect = ClientError(
            error_content, "create_thing_type"
        )
        result = boto_iot.create_thing_type(
            thingTypeName=thing_type_name,
            thingTypeDescription=thing_type_desc,
            searchableAttributesList=[thing_type_attr_1],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("create_thing_type"),
        )

    def test_that_when_deprecating_a_thing_type_succeeds_the_deprecate_thing_type_method_returns_true(
        self,
    ):
        """
        tests True when deprecate thing type succeeds
        """
        self.conn.deprecate_thing_type.return_value = {}
        result = boto_iot.deprecate_thing_type(
            thingTypeName=thing_type_name, undoDeprecate=False, **conn_parameters
        )

        self.assertTrue(result.get("deprecated"))
        self.assertEqual(result.get("error"), None)

    def test_that_when_deprecating_a_thing_type_fails_the_deprecate_thing_type_method_returns_error(
        self,
    ):
        """
        tests False when thing type fails to deprecate
        """
        self.conn.deprecate_thing_type.side_effect = ClientError(
            error_content, "deprecate_thing_type"
        )
        result = boto_iot.deprecate_thing_type(
            thingTypeName=thing_type_name, undoDeprecate=False, **conn_parameters
        )

        self.assertFalse(result.get("deprecated"))
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("deprecate_thing_type"),
        )

    def test_that_when_deleting_a_thing_type_succeeds_the_delete_thing_type_method_returns_true(
        self,
    ):
        """
        tests True when delete thing type succeeds
        """
        self.conn.delete_thing_type.return_value = {}
        result = boto_iot.delete_thing_type(
            thingTypeName=thing_type_name, **conn_parameters
        )

        self.assertTrue(result.get("deleted"))
        self.assertEqual(result.get("error"), None)

    def test_that_when_deleting_a_thing_type_fails_the_delete_thing_type_method_returns_error(
        self,
    ):
        """
        tests False when delete thing type fails
        """
        self.conn.delete_thing_type.side_effect = ClientError(
            error_content, "delete_thing_type"
        )
        result = boto_iot.delete_thing_type(
            thingTypeName=thing_type_name, **conn_parameters
        )
        self.assertFalse(result.get("deleted"))
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("delete_thing_type"),
        )


class BotoIoTPolicyTestCase(BotoIoTTestCaseBase, BotoIoTTestCaseMixin):
    """
    TestCase for salt.modules.boto_iot module
    """

    def test_that_when_checking_if_a_policy_exists_and_a_policy_exists_the_policy_exists_method_returns_true(
        self,
    ):
        """
        Tests checking iot policy existence when the iot policy already exists
        """
        self.conn.get_policy.return_value = {"policy": policy_ret}
        result = boto_iot.policy_exists(
            policyName=policy_ret["policyName"], **conn_parameters
        )

        self.assertTrue(result["exists"])

    def test_that_when_checking_if_a_policy_exists_and_a_policy_does_not_exist_the_policy_exists_method_returns_false(
        self,
    ):
        """
        Tests checking iot policy existence when the iot policy does not exist
        """
        self.conn.get_policy.side_effect = not_found_error
        result = boto_iot.policy_exists(policyName="mypolicy", **conn_parameters)

        self.assertFalse(result["exists"])

    def test_that_when_checking_if_a_policy_exists_and_boto3_returns_an_error_the_policy_exists_method_returns_error(
        self,
    ):
        """
        Tests checking iot policy existence when boto returns an error
        """
        self.conn.get_policy.side_effect = ClientError(error_content, "get_policy")
        result = boto_iot.policy_exists(policyName="mypolicy", **conn_parameters)

        self.assertEqual(
            result.get("error", {}).get("message"), error_message.format("get_policy")
        )

    def test_that_when_creating_a_policy_succeeds_the_create_policy_method_returns_true(
        self,
    ):
        """
        tests True policy created.
        """
        self.conn.create_policy.return_value = policy_ret
        result = boto_iot.create_policy(
            policyName=policy_ret["policyName"],
            policyDocument=policy_ret["policyDocument"],
            **conn_parameters
        )

        self.assertTrue(result["created"])

    def test_that_when_creating_a_policy_fails_the_create_policy_method_returns_error(
        self,
    ):
        """
        tests False policy not created.
        """
        self.conn.create_policy.side_effect = ClientError(
            error_content, "create_policy"
        )
        result = boto_iot.create_policy(
            policyName=policy_ret["policyName"],
            policyDocument=policy_ret["policyDocument"],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("create_policy"),
        )

    def test_that_when_deleting_a_policy_succeeds_the_delete_policy_method_returns_true(
        self,
    ):
        """
        tests True policy deleted.
        """
        result = boto_iot.delete_policy(policyName="testpolicy", **conn_parameters)

        self.assertTrue(result["deleted"])

    def test_that_when_deleting_a_policy_fails_the_delete_policy_method_returns_false(
        self,
    ):
        """
        tests False policy not deleted.
        """
        self.conn.delete_policy.side_effect = ClientError(
            error_content, "delete_policy"
        )
        result = boto_iot.delete_policy(policyName="testpolicy", **conn_parameters)
        self.assertFalse(result["deleted"])

    def test_that_when_describing_policy_it_returns_the_dict_of_properties_returns_true(
        self,
    ):
        """
        Tests describing parameters if policy exists
        """
        self.conn.get_policy.return_value = {"policy": policy_ret}

        result = boto_iot.describe_policy(
            policyName=policy_ret["policyName"], **conn_parameters
        )

        self.assertTrue(result["policy"])

    def test_that_when_describing_policy_it_returns_the_dict_of_properties_returns_false(
        self,
    ):
        """
        Tests describing parameters if policy does not exist
        """
        self.conn.get_policy.side_effect = not_found_error
        result = boto_iot.describe_policy(policyName="testpolicy", **conn_parameters)

        self.assertFalse(result["policy"])

    def test_that_when_describing_policy_on_client_error_it_returns_error(self):
        """
        Tests describing parameters failure
        """
        self.conn.get_policy.side_effect = ClientError(error_content, "get_policy")
        result = boto_iot.describe_policy(policyName="testpolicy", **conn_parameters)
        self.assertTrue("error" in result)

    def test_that_when_checking_if_a_policy_version_exists_and_a_policy_version_exists_the_policy_version_exists_method_returns_true(
        self,
    ):
        """
        Tests checking iot policy existence when the iot policy version already exists
        """
        self.conn.get_policy.return_value = {"policy": policy_ret}
        result = boto_iot.policy_version_exists(
            policyName=policy_ret["policyName"], policyVersionId=1, **conn_parameters
        )

        self.assertTrue(result["exists"])

    def test_that_when_checking_if_a_policy_version_exists_and_a_policy_version_does_not_exist_the_policy_version_exists_method_returns_false(
        self,
    ):
        """
        Tests checking iot policy_version existence when the iot policy_version does not exist
        """
        self.conn.get_policy_version.side_effect = not_found_error
        result = boto_iot.policy_version_exists(
            policyName=policy_ret["policyName"], policyVersionId=1, **conn_parameters
        )

        self.assertFalse(result["exists"])

    def test_that_when_checking_if_a_policy_version_exists_and_boto3_returns_an_error_the_policy_version_exists_method_returns_error(
        self,
    ):
        """
        Tests checking iot policy_version existence when boto returns an error
        """
        self.conn.get_policy_version.side_effect = ClientError(
            error_content, "get_policy_version"
        )
        result = boto_iot.policy_version_exists(
            policyName=policy_ret["policyName"], policyVersionId=1, **conn_parameters
        )

        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("get_policy_version"),
        )

    def test_that_when_creating_a_policy_version_succeeds_the_create_policy_version_method_returns_true(
        self,
    ):
        """
        tests True policy_version created.
        """
        self.conn.create_policy_version.return_value = policy_ret
        result = boto_iot.create_policy_version(
            policyName=policy_ret["policyName"],
            policyDocument=policy_ret["policyDocument"],
            **conn_parameters
        )

        self.assertTrue(result["created"])

    def test_that_when_creating_a_policy_version_fails_the_create_policy_version_method_returns_error(
        self,
    ):
        """
        tests False policy_version not created.
        """
        self.conn.create_policy_version.side_effect = ClientError(
            error_content, "create_policy_version"
        )
        result = boto_iot.create_policy_version(
            policyName=policy_ret["policyName"],
            policyDocument=policy_ret["policyDocument"],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("create_policy_version"),
        )

    def test_that_when_deleting_a_policy_version_succeeds_the_delete_policy_version_method_returns_true(
        self,
    ):
        """
        tests True policy_version deleted.
        """
        result = boto_iot.delete_policy_version(
            policyName="testpolicy", policyVersionId=1, **conn_parameters
        )

        self.assertTrue(result["deleted"])

    def test_that_when_deleting_a_policy_version_fails_the_delete_policy_version_method_returns_false(
        self,
    ):
        """
        tests False policy_version not deleted.
        """
        self.conn.delete_policy_version.side_effect = ClientError(
            error_content, "delete_policy_version"
        )
        result = boto_iot.delete_policy_version(
            policyName="testpolicy", policyVersionId=1, **conn_parameters
        )
        self.assertFalse(result["deleted"])

    def test_that_when_describing_policy_version_it_returns_the_dict_of_properties_returns_true(
        self,
    ):
        """
        Tests describing parameters if policy_version exists
        """
        self.conn.get_policy_version.return_value = {"policy": policy_ret}

        result = boto_iot.describe_policy_version(
            policyName=policy_ret["policyName"], policyVersionId=1, **conn_parameters
        )

        self.assertTrue(result["policy"])

    def test_that_when_describing_policy_version_it_returns_the_dict_of_properties_returns_false(
        self,
    ):
        """
        Tests describing parameters if policy_version does not exist
        """
        self.conn.get_policy_version.side_effect = not_found_error
        result = boto_iot.describe_policy_version(
            policyName=policy_ret["policyName"], policyVersionId=1, **conn_parameters
        )

        self.assertFalse(result["policy"])

    def test_that_when_describing_policy_version_on_client_error_it_returns_error(self):
        """
        Tests describing parameters failure
        """
        self.conn.get_policy_version.side_effect = ClientError(
            error_content, "get_policy_version"
        )
        result = boto_iot.describe_policy_version(
            policyName=policy_ret["policyName"], policyVersionId=1, **conn_parameters
        )
        self.assertTrue("error" in result)

    def test_that_when_listing_policies_succeeds_the_list_policies_method_returns_true(
        self,
    ):
        """
        tests True policies listed.
        """
        self.conn.list_policies.return_value = {"policies": [policy_ret]}
        result = boto_iot.list_policies(**conn_parameters)

        self.assertTrue(result["policies"])

    def test_that_when_listing_policy_fails_the_list_policy_method_returns_false(self):
        """
        tests False no policy listed.
        """
        self.conn.list_policies.return_value = {"policies": []}
        result = boto_iot.list_policies(**conn_parameters)
        self.assertFalse(result["policies"])

    def test_that_when_listing_policy_fails_the_list_policy_method_returns_error(self):
        """
        tests False policy error.
        """
        self.conn.list_policies.side_effect = ClientError(
            error_content, "list_policies"
        )
        result = boto_iot.list_policies(**conn_parameters)
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("list_policies"),
        )

    def test_that_when_listing_policy_versions_succeeds_the_list_policy_versions_method_returns_true(
        self,
    ):
        """
        tests True policy versions listed.
        """
        self.conn.list_policy_versions.return_value = {"policyVersions": [policy_ret]}
        result = boto_iot.list_policy_versions(
            policyName="testpolicy", **conn_parameters
        )

        self.assertTrue(result["policyVersions"])

    def test_that_when_listing_policy_versions_fails_the_list_policy_versions_method_returns_false(
        self,
    ):
        """
        tests False no policy versions listed.
        """
        self.conn.list_policy_versions.return_value = {"policyVersions": []}
        result = boto_iot.list_policy_versions(
            policyName="testpolicy", **conn_parameters
        )
        self.assertFalse(result["policyVersions"])

    def test_that_when_listing_policy_versions_fails_the_list_policy_versions_method_returns_error(
        self,
    ):
        """
        tests False policy versions error.
        """
        self.conn.list_policy_versions.side_effect = ClientError(
            error_content, "list_policy_versions"
        )
        result = boto_iot.list_policy_versions(
            policyName="testpolicy", **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("list_policy_versions"),
        )

    def test_that_when_setting_default_policy_version_succeeds_the_set_default_policy_version_method_returns_true(
        self,
    ):
        """
        tests True policy version set.
        """
        result = boto_iot.set_default_policy_version(
            policyName="testpolicy", policyVersionId=1, **conn_parameters
        )

        self.assertTrue(result["changed"])

    def test_that_when_set_default_policy_version_fails_the_set_default_policy_version_method_returns_error(
        self,
    ):
        """
        tests False policy version error.
        """
        self.conn.set_default_policy_version.side_effect = ClientError(
            error_content, "set_default_policy_version"
        )
        result = boto_iot.set_default_policy_version(
            policyName="testpolicy", policyVersionId=1, **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("set_default_policy_version"),
        )

    def test_that_when_list_principal_policies_succeeds_the_list_principal_policies_method_returns_true(
        self,
    ):
        """
        tests True policies listed.
        """
        self.conn.list_principal_policies.return_value = {"policies": [policy_ret]}
        result = boto_iot.list_principal_policies(
            principal="us-east-1:GUID-GUID-GUID", **conn_parameters
        )

        self.assertTrue(result["policies"])

    def test_that_when_list_principal_policies_fails_the_list_principal_policies_method_returns_error(
        self,
    ):
        """
        tests False policy version error.
        """
        self.conn.list_principal_policies.side_effect = ClientError(
            error_content, "list_principal_policies"
        )
        result = boto_iot.list_principal_policies(
            principal="us-east-1:GUID-GUID-GUID", **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("list_principal_policies"),
        )

    def test_that_when_attach_principal_policy_succeeds_the_attach_principal_policy_method_returns_true(
        self,
    ):
        """
        tests True policy attached.
        """
        result = boto_iot.attach_principal_policy(
            policyName="testpolicy",
            principal="us-east-1:GUID-GUID-GUID",
            **conn_parameters
        )

        self.assertTrue(result["attached"])

    def test_that_when_attach_principal_policy_version_fails_the_attach_principal_policy_version_method_returns_error(
        self,
    ):
        """
        tests False policy version error.
        """
        self.conn.attach_principal_policy.side_effect = ClientError(
            error_content, "attach_principal_policy"
        )
        result = boto_iot.attach_principal_policy(
            policyName="testpolicy",
            principal="us-east-1:GUID-GUID-GUID",
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("attach_principal_policy"),
        )

    def test_that_when_detach_principal_policy_succeeds_the_detach_principal_policy_method_returns_true(
        self,
    ):
        """
        tests True policy detached.
        """
        result = boto_iot.detach_principal_policy(
            policyName="testpolicy",
            principal="us-east-1:GUID-GUID-GUID",
            **conn_parameters
        )

        self.assertTrue(result["detached"])

    def test_that_when_detach_principal_policy_version_fails_the_detach_principal_policy_version_method_returns_error(
        self,
    ):
        """
        tests False policy version error.
        """
        self.conn.detach_principal_policy.side_effect = ClientError(
            error_content, "detach_principal_policy"
        )
        result = boto_iot.detach_principal_policy(
            policyName="testpolicy",
            principal="us-east-1:GUID-GUID-GUID",
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("detach_principal_policy"),
        )


@skipIf(HAS_BOTO is False, "The boto module must be installed.")
@skipIf(
    _has_required_boto() is False,
    "The boto3 module must be greater than"
    " or equal to version {}.  The botocore"
    " module must be greater than or equal to"
    " version {}.".format(required_boto3_version, required_botocore_version),
)
class BotoIoTTopicRuleTestCase(BotoIoTTestCaseBase, BotoIoTTestCaseMixin):
    """
    TestCase for salt.modules.boto_iot module
    """

    def test_that_when_checking_if_a_topic_rule_exists_and_a_topic_rule_exists_the_topic_rule_exists_method_returns_true(
        self,
    ):
        """
        Tests checking iot topic_rule existence when the iot topic_rule already exists
        """
        self.conn.get_topic_rule.return_value = {"rule": topic_rule_ret}
        result = boto_iot.topic_rule_exists(
            ruleName=topic_rule_ret["ruleName"], **conn_parameters
        )

        self.assertTrue(result["exists"])

    def test_that_when_checking_if_a_rule_exists_and_a_rule_does_not_exist_the_topic_rule_exists_method_returns_false(
        self,
    ):
        """
        Tests checking iot rule existence when the iot rule does not exist
        """
        self.conn.get_topic_rule.side_effect = topic_rule_not_found_error
        result = boto_iot.topic_rule_exists(ruleName="mypolicy", **conn_parameters)

        self.assertFalse(result["exists"])

    def test_that_when_checking_if_a_topic_rule_exists_and_boto3_returns_an_error_the_topic_rule_exists_method_returns_error(
        self,
    ):
        """
        Tests checking iot topic_rule existence when boto returns an error
        """
        self.conn.get_topic_rule.side_effect = ClientError(
            error_content, "get_topic_rule"
        )
        result = boto_iot.topic_rule_exists(ruleName="myrule", **conn_parameters)

        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("get_topic_rule"),
        )

    def test_that_when_creating_a_topic_rule_succeeds_the_create_topic_rule_method_returns_true(
        self,
    ):
        """
        tests True topic_rule created.
        """
        self.conn.create_topic_rule.return_value = topic_rule_ret
        result = boto_iot.create_topic_rule(
            ruleName=topic_rule_ret["ruleName"],
            sql=topic_rule_ret["sql"],
            description=topic_rule_ret["description"],
            actions=topic_rule_ret["actions"],
            **conn_parameters
        )

        self.assertTrue(result["created"])

    def test_that_when_creating_a_topic_rule_fails_the_create_topic_rule_method_returns_error(
        self,
    ):
        """
        tests False topic_rule not created.
        """
        self.conn.create_topic_rule.side_effect = ClientError(
            error_content, "create_topic_rule"
        )
        result = boto_iot.create_topic_rule(
            ruleName=topic_rule_ret["ruleName"],
            sql=topic_rule_ret["sql"],
            description=topic_rule_ret["description"],
            actions=topic_rule_ret["actions"],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("create_topic_rule"),
        )

    def test_that_when_replacing_a_topic_rule_succeeds_the_replace_topic_rule_method_returns_true(
        self,
    ):
        """
        tests True topic_rule replaced.
        """
        self.conn.replace_topic_rule.return_value = topic_rule_ret
        result = boto_iot.replace_topic_rule(
            ruleName=topic_rule_ret["ruleName"],
            sql=topic_rule_ret["sql"],
            description=topic_rule_ret["description"],
            actions=topic_rule_ret["actions"],
            **conn_parameters
        )

        self.assertTrue(result["replaced"])

    def test_that_when_replacing_a_topic_rule_fails_the_replace_topic_rule_method_returns_error(
        self,
    ):
        """
        tests False topic_rule not replaced.
        """
        self.conn.replace_topic_rule.side_effect = ClientError(
            error_content, "replace_topic_rule"
        )
        result = boto_iot.replace_topic_rule(
            ruleName=topic_rule_ret["ruleName"],
            sql=topic_rule_ret["sql"],
            description=topic_rule_ret["description"],
            actions=topic_rule_ret["actions"],
            **conn_parameters
        )
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("replace_topic_rule"),
        )

    def test_that_when_deleting_a_topic_rule_succeeds_the_delete_topic_rule_method_returns_true(
        self,
    ):
        """
        tests True topic_rule deleted.
        """
        result = boto_iot.delete_topic_rule(ruleName="testrule", **conn_parameters)

        self.assertTrue(result["deleted"])

    def test_that_when_deleting_a_topic_rule_fails_the_delete_topic_rule_method_returns_false(
        self,
    ):
        """
        tests False topic_rule not deleted.
        """
        self.conn.delete_topic_rule.side_effect = ClientError(
            error_content, "delete_topic_rule"
        )
        result = boto_iot.delete_topic_rule(ruleName="testrule", **conn_parameters)
        self.assertFalse(result["deleted"])

    def test_that_when_describing_topic_rule_it_returns_the_dict_of_properties_returns_true(
        self,
    ):
        """
        Tests describing parameters if topic_rule exists
        """
        self.conn.get_topic_rule.return_value = {"rule": topic_rule_ret}

        result = boto_iot.describe_topic_rule(
            ruleName=topic_rule_ret["ruleName"], **conn_parameters
        )

        self.assertTrue(result["rule"])

    def test_that_when_describing_topic_rule_on_client_error_it_returns_error(self):
        """
        Tests describing parameters failure
        """
        self.conn.get_topic_rule.side_effect = ClientError(
            error_content, "get_topic_rule"
        )
        result = boto_iot.describe_topic_rule(ruleName="testrule", **conn_parameters)
        self.assertTrue("error" in result)

    def test_that_when_listing_topic_rules_succeeds_the_list_topic_rules_method_returns_true(
        self,
    ):
        """
        tests True topic_rules listed.
        """
        self.conn.list_topic_rules.return_value = {"rules": [topic_rule_ret]}
        result = boto_iot.list_topic_rules(**conn_parameters)

        self.assertTrue(result["rules"])

    def test_that_when_listing_topic_rules_fails_the_list_topic_rules_method_returns_error(
        self,
    ):
        """
        tests False policy error.
        """
        self.conn.list_topic_rules.side_effect = ClientError(
            error_content, "list_topic_rules"
        )
        result = boto_iot.list_topic_rules(**conn_parameters)
        self.assertEqual(
            result.get("error", {}).get("message"),
            error_message.format("list_topic_rules"),
        )



"""
Integration tests for the vault execution module
"""

import logging
import time

import pytest
import salt.utils.path
from tests.support.case import ModuleCase
from tests.support.runtests import RUNTIME_VARS
from tests.support.sminion import create_sminion
from tests.support.unit import SkipTest, skipIf

log = logging.getLogger(__name__)

VAULT_BINARY_PATH = salt.utils.path.which("vault")


@skipIf(not salt.utils.path.which("dockerd"), "Docker not installed")
@skipIf(not VAULT_BINARY_PATH, "Vault not installed")
@pytest.mark.destructive_test
class VaultTestCase(ModuleCase):
    """
    Test vault module
    """

    @classmethod
    def setUpClass(cls):
        cls.sminion = sminion = create_sminion()
        config = (
            '{"backend": {"file": {"path": "/vault/file"}}, "default_lease_ttl":'
            ' "168h", "max_lease_ttl": "720h", "disable_mlock": true}'
        )
        sminion.states.docker_image.present(name="vault", tag="0.9.6")
        login_attempts = 1
        container_created = False
        while True:
            if container_created:
                sminion.states.docker_container.stopped(name="vault")
                sminion.states.docker_container.absent(name="vault")
            ret = sminion.states.docker_container.running(
                name="vault",
                image="vault:0.9.6",
                port_bindings="8200:8200",
                environment={
                    "VAULT_DEV_ROOT_TOKEN_ID": "testsecret",
                    "VAULT_LOCAL_CONFIG": config,
                },
            )
            log.debug("docker_container.running return: %s", ret)
            container_created = ret["result"]
            time.sleep(5)
            ret = sminion.functions.cmd.run_all(
                cmd="{} login token=testsecret".format(VAULT_BINARY_PATH),
                env={"VAULT_ADDR": "http://127.0.0.1:8200"},
                hide_output=False,
            )
            if ret["retcode"] == 0:
                break
            log.debug("Vault login failed. Return: %s", ret)
            login_attempts += 1

            if login_attempts >= 3:
                raise SkipTest("unable to login to vault")

        ret = sminion.functions.cmd.retcode(
            cmd="{} policy write testpolicy {}/vault.hcl".format(
                VAULT_BINARY_PATH, RUNTIME_VARS.FILES
            ),
            env={"VAULT_ADDR": "http://127.0.0.1:8200"},
        )
        if ret != 0:
            raise SkipTest("unable to assign policy to vault")

    @classmethod
    def tearDownClass(cls):
        cls.sminion.states.docker_container.stopped(name="vault")
        cls.sminion.states.docker_container.absent(name="vault")
        cls.sminion.states.docker_image.absent(name="vault", force=True)
        cls.sminion = None

    @pytest.mark.slow_test
    def test_write_read_secret(self):
        write_return = self.run_function(
            "vault.write_secret", path="secret/my/secret", user="foo", password="bar"
        )
        self.assertEqual(write_return, True)
        assert self.run_function("vault.read_secret", arg=["secret/my/secret"]) == {
            "password": "bar",
            "user": "foo",
        }
        assert (
            self.run_function("vault.read_secret", arg=["secret/my/secret", "user"])
            == "foo"
        )

    @pytest.mark.slow_test
    def test_write_raw_read_secret(self):
        assert (
            self.run_function(
                "vault.write_raw",
                path="secret/my/secret2",
                raw={"user2": "foo2", "password2": "bar2"},
            )
            is True
        )
        assert self.run_function("vault.read_secret", arg=["secret/my/secret2"]) == {
            "password2": "bar2",
            "user2": "foo2",
        }

    @pytest.mark.slow_test
    def test_delete_secret(self):
        assert (
            self.run_function(
                "vault.write_secret",
                path="secret/my/secret",
                user="foo",
                password="bar",
            )
            is True
        )
        assert (
            self.run_function("vault.delete_secret", arg=["secret/my/secret"]) is True
        )

    @pytest.mark.slow_test
    def test_list_secrets(self):
        assert (
            self.run_function(
                "vault.write_secret",
                path="secret/my/secret",
                user="foo",
                password="bar",
            )
            is True
        )
        assert self.run_function("vault.list_secrets", arg=["secret/my/"]) == {
            "keys": ["secret"]
        }


@skipIf(not salt.utils.path.which("dockerd"), "Docker not installed")
@skipIf(not salt.utils.path.which("vault"), "Vault not installed")
@pytest.mark.destructive_test
class VaultTestCaseCurrent(ModuleCase):
    """
    Test vault module against current vault
    """

    @classmethod
    def setUpClass(cls):
        cls.sminion = sminion = create_sminion()
        config = (
            '{"backend": {"file": {"path": "/vault/file"}}, "default_lease_ttl":'
            ' "168h", "max_lease_ttl": "720h", "disable_mlock": true}'
        )
        sminion.states.docker_image.present(name="vault", tag="1.3.1")
        login_attempts = 1
        container_created = False
        while True:
            if container_created:
                sminion.states.docker_container.stopped(name="vault")
                sminion.states.docker_container.absent(name="vault")
            ret = sminion.states.docker_container.running(
                name="vault",
                image="vault:1.3.1",
                port_bindings="8200:8200",
                environment={
                    "VAULT_DEV_ROOT_TOKEN_ID": "testsecret",
                    "VAULT_LOCAL_CONFIG": config,
                },
            )
            log.debug("docker_container.running return: %s", ret)
            container_created = ret["result"]
            time.sleep(5)
            ret = sminion.functions.cmd.run_all(
                cmd="{} login token=testsecret".format(VAULT_BINARY_PATH),
                env={"VAULT_ADDR": "http://127.0.0.1:8200"},
                hide_output=False,
            )
            if ret["retcode"] == 0:
                break
            log.debug("Vault login failed. Return: %s", ret)
            login_attempts += 1

            if login_attempts >= 3:
                raise SkipTest("unable to login to vault")

        ret = sminion.functions.cmd.retcode(
            cmd="{} policy write testpolicy {}/vault.hcl".format(
                VAULT_BINARY_PATH, RUNTIME_VARS.FILES
            ),
            env={"VAULT_ADDR": "http://127.0.0.1:8200"},
        )
        if ret != 0:
            raise SkipTest("unable to assign policy to vault")

    @classmethod
    def tearDownClass(cls):
        cls.sminion.states.docker_container.stopped(name="vault")
        cls.sminion.states.docker_container.absent(name="vault")
        cls.sminion.states.docker_image.absent(name="vault", force=True)
        cls.sminion = None

    @pytest.mark.slow_test
    def test_write_read_secret_kv2(self):
        write_return = self.run_function(
            "vault.write_secret", path="secret/my/secret", user="foo", password="bar"
        )
        # write_secret output:
        # {'created_time': '2020-01-12T23:09:34.571294241Z', 'destroyed': False,
        # 'version': 1, 'deletion_time': ''}
        expected_write = {"destroyed": False, "deletion_time": ""}
        self.assertDictContainsSubset(expected_write, write_return)

        read_return = self.run_function(
            "vault.read_secret", arg=["secret/my/secret"], metadata=True
        )
        # read_secret output:
        # {'data': {'password': 'bar', 'user': 'foo'},
        # 'metadata': {'created_time': '2020-01-12T23:07:18.829326918Z', 'destroyed': False,
        # 'version': 1, 'deletion_time': ''}}
        expected_read = {"data": {"password": "bar", "user": "foo"}}
        self.assertDictContainsSubset(expected_read, read_return)
        expected_read = {"password": "bar", "user": "foo"}
        read_return = self.run_function("vault.read_secret", arg=["secret/my/secret"])
        self.assertDictContainsSubset(expected_read, read_return)

        read_return = self.run_function(
            "vault.read_secret", arg=["secret/my/secret", "user"]
        )
        self.assertEqual(read_return, "foo")

    @pytest.mark.slow_test
    def test_list_secrets_kv2(self):
        write_return = self.run_function(
            "vault.write_secret", path="secret/my/secret", user="foo", password="bar"
        )
        expected_write = {"destroyed": False, "deletion_time": ""}
        self.assertDictContainsSubset(expected_write, write_return)

        list_return = self.run_function("vault.list_secrets", arg=["secret/my/"])
        self.assertIn("secret", list_return["keys"])

    @pytest.mark.slow_test
    def test_write_raw_read_secret_kv2(self):
        write_return = self.run_function(
            "vault.write_raw",
            path="secret/my/secret2",
            raw={"user2": "foo2", "password2": "bar2"},
        )
        expected_write = {"destroyed": False, "deletion_time": ""}
        self.assertDictContainsSubset(expected_write, write_return)

        read_return = self.run_function(
            "vault.read_secret", arg=["secret/my/secret2"], metadata=True
        )
        expected_read = {"data": {"password2": "bar2", "user2": "foo2"}}
        self.assertDictContainsSubset(expected_read, read_return)

        read_return = self.run_function("vault.read_secret", arg=["secret/my/secret2"])
        expected_read = {"password2": "bar2", "user2": "foo2"}
        self.assertDictContainsSubset(expected_read, read_return)

    @pytest.mark.slow_test
    def test_delete_secret_kv2(self):
        write_return = self.run_function(
            "vault.write_secret",
            path="secret/my/secret3",
            user3="foo3",
            password3="bar3",
        )
        expected_write = {"destroyed": False, "deletion_time": ""}
        self.assertDictContainsSubset(expected_write, write_return)

        delete_return = self.run_function(
            "vault.delete_secret", arg=["secret/my/secret3"]
        )
        self.assertEqual(delete_return, True)

    @pytest.mark.slow_test
    def test_destroy_secret_kv2(self):
        write_return = self.run_function(
            "vault.write_secret",
            path="secret/my/secret4",
            user3="foo4",
            password4="bar4",
        )
        expected_write = {"destroyed": False, "deletion_time": ""}
        self.assertDictContainsSubset(expected_write, write_return)

        destroy_return = self.run_function(
            "vault.destroy_secret", arg=["secret/my/secret4", "1"]
        )
        self.assertEqual(destroy_return, True)
        # self.assertIsNone(self.run_function('vault.read_secret', arg=['secret/my/secret4']))
        # list_return = self.run_function('vault.list_secrets', arg=['secret/my/'])
        # self.assertNotIn('secret4', list_return['keys'])



import os
import tempfile

import salt.crypt
import salt.utils.files
from salt.cloud.clouds import ec2
from salt.exceptions import SaltCloudSystemExit
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import PropertyMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, skipIf
from tests.unit.test_crypt import PRIVKEY_DATA

PASS_DATA = (
    b"qOjCKDlBdcNEbJ/J8eRl7sH+bYIIm4cvHHY86gh2NEUnufFlFo0gGVTZR05Fj0cw3n/w7gR"
    b"urNXz5JoeSIHVuNI3YTwzL9yEAaC0kuy8EbOlO2yx8yPGdfml9BRwOV7A6b8UFo9co4H7fz"
    b"DdScMKU2yzvRYvp6N6Q2cJGBmPsemnXWWusb+1vZVWxcRAQmG3ogF6Z5rZSYAYH0N4rqJgH"
    b"mQfzuyb+jrBvV/IOoV1EdO9jGSH9338aS47NjrmNEN/SpnS6eCWZUwwyHbPASuOvWiY4QH/"
    b"0YZC6EGccwiUmt0ZOxIynk+tEyVPTkiS0V8RcZK6YKqMWHpKmPtLBzfuoA=="
)


class EC2TestCase(TestCase, LoaderModuleMockMixin):
    """
    Unit TestCase for salt.cloud.clouds.ec2 module.
    """

    def setUp(self):
        super().setUp()
        with tempfile.NamedTemporaryFile(
            dir=RUNTIME_VARS.TMP, suffix=".pem", delete=True
        ) as fp:
            self.key_file = fp.name

    def tearDown(self):
        super().tearDown()
        if os.path.exists(self.key_file):
            os.remove(self.key_file)

    def setup_loader_modules(self):
        return {ec2: {"__opts__": {}}}

    def test__validate_key_path_and_mode(self):
        # Key file exists
        with patch("os.path.exists", return_value=True):
            with patch("os.stat") as patched_stat:
                type(patched_stat.return_value).st_mode = PropertyMock(
                    return_value=0o644
                )
                self.assertRaises(
                    SaltCloudSystemExit, ec2._validate_key_path_and_mode, "key_file"
                )

                type(patched_stat.return_value).st_mode = PropertyMock(
                    return_value=0o600
                )
                self.assertTrue(ec2._validate_key_path_and_mode("key_file"))

                type(patched_stat.return_value).st_mode = PropertyMock(
                    return_value=0o400
                )
                self.assertTrue(ec2._validate_key_path_and_mode("key_file"))

        # Key file does not exist
        with patch("os.path.exists", return_value=False):
            self.assertRaises(
                SaltCloudSystemExit, ec2._validate_key_path_and_mode, "key_file"
            )

    @skipIf(not salt.crypt.HAS_M2 and not salt.crypt.HAS_CRYPTO, "Needs crypto library")
    @patch("salt.cloud.clouds.ec2._get_node")
    @patch("salt.cloud.clouds.ec2.get_location")
    @patch("salt.cloud.clouds.ec2.get_provider")
    @patch("salt.utils.aws.query")
    def test_get_password_data(self, query, get_provider, get_location, _get_node):
        query.return_value = [{"passwordData": PASS_DATA}]
        _get_node.return_value = {"instanceId": "i-abcdef"}
        get_location.return_value = "us-west2"
        get_provider.return_value = "ec2"
        with salt.utils.files.fopen(self.key_file, "w") as fp:
            fp.write(PRIVKEY_DATA)
        ret = ec2.get_password_data(
            name="i-abcddef", kwargs={"key_file": self.key_file}, call="action"
        )
        assert ret["passwordData"] == PASS_DATA
        assert ret["password"] == "testp4ss!"

    @patch("salt.cloud.clouds.ec2.config.get_cloud_config_value")
    @patch("salt.cloud.clouds.ec2.get_location")
    @patch("salt.cloud.clouds.ec2.get_provider")
    @patch("salt.cloud.clouds.ec2.aws.query")
    def test_get_imageid(self, aws_query, get_provider, get_location, config):
        """
        test querying imageid function
        """
        vm = {}
        ami = "ami-1234"
        config.return_value = "test/*"
        get_location.return_value = "us-west2"
        get_provider.return_value = "ec2"
        aws_query.return_value = [{"imageId": ami}]

        # test image filter
        self.assertEqual(ec2.get_imageid(vm), ami)

        # test ami-image
        config.return_value = ami
        self.assertEqual(ec2.get_imageid(vm), ami)

        # we should have only ran aws.query once when testing the aws filter
        aws_query.assert_called_once()

    @patch("salt.cloud.clouds.ec2.config.get_cloud_config_value")
    @patch("salt.cloud.clouds.ec2.get_location")
    @patch("salt.cloud.clouds.ec2.get_availability_zone")
    @patch("salt.cloud.clouds.ec2.get_provider")
    @patch("salt.cloud.clouds.ec2.get_spot_config")
    @patch("salt.cloud.clouds.ec2._param_from_config")
    @patch("salt.cloud.clouds.ec2.securitygroupid")
    def test_termination_protection(
        self,
        securitygroupid,
        _param_from_config,
        get_spot_config,
        get_provider,
        get_availability_zone,
        get_location,
        config,
    ):
        """
        Verify that `set_termination_protection` updates the right parameters
        """
        vm = {"name": "taco"}
        set_del_root_vol_on_destroy = "yes"
        termination_protection = True
        config.side_effect = (
            [None] * 2
            + ["test/*"]
            + [None] * 13
            + [set_del_root_vol_on_destroy, termination_protection]
        )
        get_location.return_value = "us-west2"
        get_availability_zone.return_value = None
        get_provider.return_value = "ec2"
        get_spot_config.return_value = None
        securitygroupid.return_value = None

        self.assertRaises(
            salt.exceptions.SaltCloudConfigError, ec2.request_instance, vm
        )
        _param_from_config.assert_called_once_with("DisableApiTermination", True)

    @patch("salt.cloud.clouds.ec2.config.get_cloud_config_value")
    @patch("salt.cloud.clouds.ec2.get_location")
    @patch("salt.cloud.clouds.ec2.get_availability_zone")
    @patch("salt.cloud.clouds.ec2.get_provider")
    @patch("salt.cloud.clouds.ec2.get_spot_config")
    @patch("salt.cloud.clouds.ec2.securitygroupid")
    def test_termination_protection_exception(
        self,
        securitygroupid,
        get_spot_config,
        get_provider,
        get_availability_zone,
        get_location,
        config,
    ):
        """
        Verify improper `set_termination_protection` parameters raises an exception
        """
        vm = {"name": "taco"}
        termination_protection = "not a bool"
        config.side_effect = (
            [None] * 2 + ["test/*"] + [None] * 14 + [termination_protection]
        )
        get_location.return_value = "us-west2"
        get_availability_zone.return_value = None
        get_provider.return_value = "ec2"
        get_spot_config.return_value = None
        securitygroupid.return_value = None

        self.assertRaises(
            salt.exceptions.SaltCloudConfigError, ec2.request_instance, vm
        )

    @patch("salt.cloud.clouds.ec2.config.get_cloud_config_value")
    @patch("salt.cloud.clouds.ec2.get_location")
    @patch("salt.cloud.clouds.ec2.get_provider")
    @patch("salt.cloud.clouds.ec2.aws.query")
    def test_get_subnetname_id(self, aws_query, get_provider, get_location, config):
        """
        test querying subnetid function
        """
        vm = {}
        subnetid = "subnet-5678"
        subnetname = "valid-subnet-with-name"
        config.return_value = subnetname
        get_location.return_value = "us-west-2"
        get_provider.return_value = "ec2"

        # test for returns that include subnets with missing Name tags, see Issue 44330
        aws_query.return_value = [
            {"subnetId": "subnet-1234"},
            {
                "subnetId": subnetid,
                "tagSet": {"item": {"key": "Name", "value": subnetname}},
            },
        ]

        # test subnetname lookup
        self.assertEqual(ec2._get_subnetname_id(subnetname), subnetid)



"""
tests.unit.returners.local_cache_test
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the Default Job Cache (local_cache).
"""

import logging
import os
import shutil
import tempfile
import time

import pytest
import salt.returners.local_cache as local_cache
import salt.utils.files
import salt.utils.jid
import salt.utils.job
import salt.utils.platform
from tests.support.mixins import (
    AdaptedConfigurationTestCaseMixin,
    LoaderModuleMockMixin,
)
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class LocalCacheCleanOldJobsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Tests for the local_cache.clean_old_jobs function.
    """

    @classmethod
    def setUpClass(cls):
        cls.TMP_CACHE_DIR = tempfile.mkdtemp(
            prefix="salt_test_job_cache", dir=RUNTIME_VARS.TMP
        )
        cls.TMP_JID_DIR = os.path.join(cls.TMP_CACHE_DIR, "jobs")

    def setup_loader_modules(self):
        return {
            local_cache: {"__opts__": {"cachedir": self.TMP_CACHE_DIR, "keep_jobs": 1}}
        }

    def tearDown(self):
        """
        Clean up after tests.

        Note that a setUp function is not used in this TestCase because the
        _make_tmp_jid_dirs replaces it.
        """
        if os.path.exists(self.TMP_CACHE_DIR):
            shutil.rmtree(self.TMP_CACHE_DIR, ignore_errors=True)

    def test_clean_old_jobs_no_jid_root(self):
        """
        Tests that the function returns None when no jid_root is found.
        """
        with patch("os.path.exists", MagicMock(return_value=False)):
            self.assertEqual(local_cache.clean_old_jobs(), None)

    def test_clean_old_jobs_empty_jid_dir_removed(self):
        """
        Tests that an empty JID dir is removed when it is old enough to be deleted.
        """
        # Create temp job cache dir without files in it.
        jid_dir, jid_file = self._make_tmp_jid_dirs(create_files=False)

        # File timestamps on Windows aren't as precise. Let a little time pass
        if salt.utils.platform.is_windows():
            time.sleep(0.01)

        # Make sure there are no files in the directory before continuing
        self.assertEqual(jid_file, None)

        # Call clean_old_jobs function, patching the keep_jobs value with a
        # very small value to force the call to clean the job.
        with patch.dict(local_cache.__opts__, {"keep_jobs": 0.00000001}):
            # Sleep on Windows because time.time is only precise to 3 decimal
            # points, and therefore subtracting the jid_ctime from time.time
            # will result in a negative number
            if salt.utils.platform.is_windows():
                time.sleep(0.25)
            local_cache.clean_old_jobs()

        # Assert that the JID dir was removed
        self.assertEqual([], os.listdir(self.TMP_JID_DIR))

    def test_clean_old_jobs_empty_jid_dir_remains(self):
        """
        Tests that an empty JID dir is NOT removed because it was created within
        the keep_jobs time frame.
        """
        # Create temp job cache dir without files in it.
        jid_dir, jid_file = self._make_tmp_jid_dirs(create_files=False)

        # Make sure there are no files in the directory
        self.assertEqual(jid_file, None)

        # Call clean_old_jobs function
        local_cache.clean_old_jobs()

        # Get the name of the JID directory that was created to test against
        if salt.utils.platform.is_windows():
            jid_dir_name = jid_dir.rpartition("\\")[2]
        else:
            jid_dir_name = jid_dir.rpartition("/")[2]

        # Assert the JID directory is still present to be cleaned after keep_jobs interval
        self.assertEqual([jid_dir_name], os.listdir(self.TMP_JID_DIR))

    def test_clean_old_jobs_jid_file_corrupted(self):
        """
        Tests that the entire JID dir is removed when the jid_file is not a file.
        This scenario indicates a corrupted cache entry, so the entire dir is scrubbed.
        """
        # Create temp job cache dir and jid file
        jid_dir, jid_file = self._make_tmp_jid_dirs()

        # Make sure there is a jid file in a new job cache director
        jid_dir_name = jid_file.rpartition("/")[2]
        self.assertEqual(jid_dir_name, "jid")

        # Even though we created a valid jid file in the _make_tmp_jid_dirs call to get
        # into the correct loop, we need to mock the 'os.path.isfile' check to force the
        # "corrupted file" check in the clean_old_jobs call.
        with patch("os.path.isfile", MagicMock(return_value=False)) as mock:
            local_cache.clean_old_jobs()

        # there should be only 1 dir in TMP_JID_DIR
        self.assertEqual(1, len(os.listdir(self.TMP_JID_DIR)))
        # top level dir should still be present
        self.assertEqual(True, os.path.exists(jid_dir))
        self.assertEqual(True, os.path.isdir(jid_dir))
        # while the 'jid' dir inside it should be gone
        self.assertEqual(False, os.path.exists(jid_dir_name))

    def test_clean_old_jobs_jid_file_is_cleaned(self):
        """
        Test that the entire JID dir is removed when a job is old enough to be removed.
        """
        # Create temp job cache dir and jid file
        jid_dir, jid_file = self._make_tmp_jid_dirs()

        # File timestamps on Windows aren't as precise. Let a little time pass
        if salt.utils.platform.is_windows():
            time.sleep(0.01)

        # Make sure there is a jid directory
        jid_dir_name = jid_file.rpartition("/")[2]
        self.assertEqual(jid_dir_name, "jid")

        # Call clean_old_jobs function, patching the keep_jobs value with a
        # very small value to force the call to clean the job.
        with patch.dict(local_cache.__opts__, {"keep_jobs": 0.00000001}):
            # Sleep on Windows because time.time is only precise to 3 decimal
            # points, and therefore subtracting the jid_ctime from time.time
            # will result in a negative number
            if salt.utils.platform.is_windows():
                time.sleep(0.25)
            local_cache.clean_old_jobs()

        # there should be only 1 dir in TMP_JID_DIR
        self.assertEqual(1, len(os.listdir(self.TMP_JID_DIR)))
        # top level dir should still be present
        self.assertEqual(True, os.path.exists(jid_dir))
        self.assertEqual(True, os.path.isdir(jid_dir))
        # while the 'jid' dir inside it should be gone
        self.assertEqual(False, os.path.exists(jid_dir_name))

    def _make_tmp_jid_dirs(self, create_files=True):
        """
        Helper function to set up temporary directories and files used for
        testing the clean_old_jobs function.

        This emulates salt.utils.jid.jid_dir() by creating this structure:

        RUNTIME_VARS.TMP_JID_DIR dir/
          random dir from tempfile.mkdtemp/
            'jid' directory/
              'jid' file

        Returns a temp_dir name and a jid_file_path. If create_files is False,
        the jid_file_path will be None.
        """
        # First, create the /tmp/salt_test_job_cache/jobs/ directory to hold jid dirs
        if not os.path.exists(self.TMP_JID_DIR):
            os.makedirs(self.TMP_JID_DIR)

        # Then create a JID temp file in "/tmp/salt_test_job_cache/"
        temp_dir = tempfile.mkdtemp(dir=self.TMP_JID_DIR)

        jid_file_path = None
        if create_files:
            dir_name = "/".join([temp_dir, "jid"])
            os.mkdir(dir_name)
            jid_file_path = "/".join([dir_name, "jid"])
            with salt.utils.files.fopen(jid_file_path, "w") as jid_file:
                jid_file.write("this is a jid file")

        return temp_dir, jid_file_path


class Local_CacheTest(
    TestCase, AdaptedConfigurationTestCaseMixin, LoaderModuleMockMixin
):
    """
    Test the local cache returner
    """

    def setup_loader_modules(self):
        return {
            local_cache: {
                "__opts__": {
                    "cachedir": self.TMP_CACHE_DIR,
                    "keep_jobs": self.KEEP_JOBS,
                }
            }
        }

    @classmethod
    def setUpClass(cls):
        cls.TMP_CACHE_DIR = tempfile.mkdtemp(
            prefix="salt_test_local_cache", dir=RUNTIME_VARS.TMP
        )
        cls.JOBS_DIR = os.path.join(cls.TMP_CACHE_DIR, "jobs")
        cls.JID_DIR = os.path.join(
            cls.JOBS_DIR,
            "31",
            "c56eed380a4e899ae12bc42563cfdfc53066fb4a6b53e2378a08ac49064539",
        )
        cls.JID_FILE = os.path.join(cls.JID_DIR, "jid")
        cls.JID_MINION_DIR = os.path.join(cls.JID_DIR, "minion", "return.p")
        cls.JOB_CACHE_DIR_FILES = [cls.JID_FILE, cls.JID_MINION_DIR]
        cls.KEEP_JOBS = 0.0000000010
        cls.EMPTY_JID_DIR = []

    @classmethod
    def tearDownClass(cls):
        for attrname in (
            "TMP_CACHE_DIR",
            "JOBS_DIR",
            "JID_DIR",
            "JID_FILE",
            "JID_MINION_DIR",
            "JOB_CACHE_DIR_FILES",
            "KEEP_JOBS",
            "EMPTY_JID_DIR",
        ):
            try:
                attr_instance = getattr(cls, attrname)
                if isinstance(attr_instance, str):
                    if os.path.isdir(attr_instance):
                        shutil.rmtree(attr_instance, ignore_errors=True)
                    elif os.path.isfile(attr_instance):
                        os.unlink(attr_instance)
                delattr(cls, attrname)
            except AttributeError:
                continue

    def _check_dir_files(self, msg, contents, status="None"):
        """
        helper method to ensure files or dirs
        are either present or removed
        """
        for content in contents:
            log.debug("CONTENT %s", content)
            if status == "present":
                check_job_dir = os.path.exists(content)
            elif status == "removed":
                if os.path.exists(content):
                    check_job_dir = False
                else:
                    check_job_dir = True
            self.assertTrue(check_job_dir, msg=msg + content)

    def _add_job(self):
        """
        helper method to add job.
        """
        # add the job.
        opts = self.get_temp_config("master")
        opts["cachedir"] = self.TMP_CACHE_DIR
        load = {
            "fun_args": [],
            "jid": "20160603132323715452",
            "return": True,
            "retcode": 0,
            "success": True,
            "cmd": "_return",
            "fun": "test.ping",
            "id": "minion",
        }

        add_job = salt.utils.job.store_job(opts, load)
        self.assertEqual(add_job, None)
        self._check_dir_files(
            "Dir/file does not exist: ", self.JOB_CACHE_DIR_FILES, status="present"
        )

    @pytest.mark.slow_test
    def test_clean_old_jobs(self):
        """
        test to ensure jobs are removed from job cache
        """
        self._add_job()

        if salt.utils.platform.is_windows():
            time.sleep(0.01)

        # remove job
        self.assertEqual(local_cache.clean_old_jobs(), None)

        self._check_dir_files(
            "job cache was not removed: ", self.JOB_CACHE_DIR_FILES, status="removed"
        )

    @pytest.mark.slow_test
    def test_not_clean_new_jobs(self):
        """
        test to ensure jobs are not removed when
        jobs dir is new
        """
        self._add_job()

        with patch.dict(local_cache.__opts__, {"keep_jobs": 24}):
            self.assertEqual(local_cache.clean_old_jobs(), None)

            self._check_dir_files(
                "job cache was removed: ", self.JOB_CACHE_DIR_FILES, status="present"
            )

    @pytest.mark.slow_test
    def test_empty_jid_dir(self):
        """
        test to ensure removal of empty jid dir
        """
        # add empty jid dir
        new_jid_dir = os.path.join(self.JOBS_DIR, "z0")
        self.EMPTY_JID_DIR.append(new_jid_dir)
        os.makedirs(new_jid_dir)

        # This needed due to a race condition in Windows
        # `os.makedirs` hasn't released the handle before
        # `local_cache.clean_old_jobs` tries to delete the new_jid_dir
        if salt.utils.platform.is_windows():
            import time

            lock_dir = new_jid_dir + ".lckchk"
            tries = 0
            while True:
                tries += 1
                if tries > 10:
                    break
                # Rename the directory and name it back
                # If it fails, the directory handle is not released, try again
                # If it succeeds, break and continue test
                try:
                    os.rename(new_jid_dir, lock_dir)
                    time.sleep(1)
                    os.rename(lock_dir, new_jid_dir)
                    break
                except OSError:  # pylint: disable=E0602
                    continue

        # check dir exists
        self._check_dir_files(
            "new_jid_dir was not created", self.EMPTY_JID_DIR, status="present"
        )

        # remove job
        self.assertEqual(local_cache.clean_old_jobs(), None)

        # check jid dir is removed
        self._check_dir_files(
            "new_jid_dir was not removed", self.EMPTY_JID_DIR, status="removed"
        )
